import asyncio
import signal
import threading

from kopf._core.reactor.running import ultimate_termination


async def test_no_kill_scheduled_when_stop_flag_is_set(settings, mocker, looptime):
    pthread_kill = mocker.patch.object(signal, 'pthread_kill')
    stop_flag = asyncio.Event()
    stop_flag.set()

    task = asyncio.create_task(ultimate_termination(
        settings=settings,
        stop_flag=stop_flag,
    ))

    await asyncio.sleep(0)
    task.cancel()
    await asyncio.sleep(0)

    assert task.done()
    assert not pthread_kill.called
    assert looptime == 0


async def test_no_kill_scheduled_when_timeout_is_none(settings, mocker, looptime):
    pthread_kill = mocker.patch.object(signal, 'pthread_kill')
    settings.process.ultimate_exiting_timeout = None

    task = asyncio.create_task(ultimate_termination(
        settings=settings,
        stop_flag=None,
    ))

    await asyncio.sleep(0)
    task.cancel()
    await asyncio.sleep(0)

    assert task.done()

    # Even after waiting, no kill should be scheduled.
    await asyncio.sleep(10)

    assert not pthread_kill.called
    assert looptime == 10


async def test_kill_scheduled_when_stop_flag_is_not_set(settings, mocker, looptime):
    pthread_kill = mocker.patch.object(signal, 'pthread_kill')
    mocker.patch.object(threading, 'get_ident', return_value=12345)
    settings.process.ultimate_exiting_timeout = 10

    task = asyncio.create_task(ultimate_termination(
        settings=settings,
        stop_flag=None,
    ))

    await asyncio.sleep(0)
    task.cancel()
    await asyncio.sleep(0)

    assert task.done()
    assert not pthread_kill.called

    # Fast-forward past the scheduled call_later timeout.
    await asyncio.sleep(10)

    assert pthread_kill.call_count == 1
    assert pthread_kill.call_args.args == (12345, signal.SIGKILL)
    assert looptime == 10



"""
Conversion of low-level events to high-level causes, and handling them.

These functions are invoked from :mod:`kopf._core.reactor.queueing`,
which are the actual event loop of the operator process.

The conversion of the low-level events to the high-level causes is done by
checking the object's state and comparing it to the preserved last-seen state.

The framework itself makes the necessary changes to the object, --- such as the
finalizers attachment, last-seen state updates, and handler status tracking, ---
thus provoking the low-level watch-events and additional queueing calls.
But these internal changes are filtered out from the cause detection
and therefore do not trigger the user-defined handlers.
"""
import asyncio
import contextlib
import functools
from collections.abc import Collection
from typing import NamedTuple

from kopf._cogs.aiokits import aiotime, aiotoggles
from kopf._cogs.configs import configuration
from kopf._cogs.structs import bodies, diffs, ephemera, finalizers, patches, references
from kopf._core.actions import application, execution, lifecycles, loggers, progression, throttlers
from kopf._core.engines import daemons, indexing, posting
from kopf._core.intents import causes, registries
from kopf._core.reactor import inventory, subhandling


async def process_resource_event(
        lifecycle: execution.LifeCycleFn,
        indexers: indexing.OperatorIndexers,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        memories: inventory.ResourceMemories,
        memobase: ephemera.AnyMemo,
        resource: references.Resource,
        raw_event: bodies.RawEvent,
        event_queue: posting.K8sEventQueue,
        stream_pressure: asyncio.Event | None = None,  # None for tests
        operator_paused: aiotoggles.ToggleSet | None = None,  # None for tests & observation
        resource_indexed: aiotoggles.Toggle | None = None,  # None for tests & observation
        operator_indexed: aiotoggles.ToggleSet | None = None,  # None for tests & observation
        consistency_time: float | None = None,  # None for tests
        no_throttling: bool = False,  # for tests & simulations
) -> str | None:  # patched resource version, if patched
    """
    Handle a single custom object low-level watch-event.

    Convert the low-level events, as provided by the watching/queueing tasks,
    to the high-level causes, and then call the cause-handling logic.
    """

    # Recall what is stored about that object. Share it in little portions with the consumers.
    # And immediately forget it if the object is deleted from the cluster (but keep in memory).
    raw_type, raw_body = raw_event['type'], raw_event['object']
    memory = await memories.recall(raw_body, noticed_by_listing=raw_type is None, memobase=memobase)
    if memory.daemons_memory.live_fresh_body is not None:
        memory.daemons_memory.live_fresh_body._replace_with(raw_body)
    if raw_type == 'DELETED':
        await memories.forget(raw_body)

    # Convert to a heavy mapping-view wrapper only now, when heavy processing begins.
    # Raw-event streaming, queueing, and batching use regular lightweight dicts.
    # Why here? 1. Before it splits into multiple causes & handlers for the same object's body;
    # 2. After it is batched (queueing); 3. While the "raw" parsed JSON is still known;
    # 4. Same as where a patch object of a similar wrapping semantics is created.
    live_fresh_body = memory.daemons_memory.live_fresh_body
    body = live_fresh_body if live_fresh_body is not None else bodies.Body(raw_body)
    patch = patches.Patch(memory.remaining_patch, body=raw_body)

    # Different loggers for different cases with different verbosity and exposure.
    local_logger = loggers.LocalObjectLogger(body=body, settings=settings)
    terse_logger = loggers.TerseObjectLogger(body=body, settings=settings)
    event_logger = loggers.ObjectLogger(body=body, settings=settings)

    # Throttle the non-handler-related errors. The regular event watching/batching continues
    # to prevent queue overfilling, but the processing is skipped (events are ignored).
    # Choice of place: late enough to have a per-resource memory for a throttler; also, a logger.
    # But early enough to catch environment errors from K8s API, and from most of the complex code.
    throttled = contextlib.nullcontext(True) if no_throttling else throttlers.throttled(
        throttler=memory.error_throttler,
        logger=local_logger,
        delays=settings.queueing.error_delays,
        wakeup=stream_pressure,
    )
    async with throttled as should_run:
        if should_run:

            # Each object has its own prefixed logger, to distinguish parallel handling.
            posting.event_queue_loop_var.set(asyncio.get_running_loop())
            posting.event_queue_var.set(event_queue)  # till the end of this object's task.

            # [Pre-]populate the indices. This must be lightweight.
            await indexing.index_resource(
                registry=registry,
                indexers=indexers,
                settings=settings,
                resource=resource,
                raw_event=raw_event,
                body=body,
                memo=memory.memo,
                memory=memory.indexing_memory,
                logger=terse_logger,
            )

            # Wait for all other individual resources and all other resource kinds' lists to finish.
            if operator_indexed is not None and resource_indexed is not None:
                await operator_indexed.drop_toggle(resource_indexed)
            if operator_indexed is not None:
                await operator_indexed.wait_for(True)  # other resource kinds & objects.

            # Do the magic -- do the job.
            delays, matched = await process_resource_causes(
                lifecycle=lifecycle,
                indexers=indexers,
                registry=registry,
                settings=settings,
                resource=resource,
                raw_event=raw_event,
                body=body,
                patch=patch,
                memory=memory,
                local_logger=local_logger,
                event_logger=event_logger,
                stream_pressure=stream_pressure,
                operator_paused=operator_paused,
                consistency_time=consistency_time,
            )

            # Whatever was done, apply the accumulated changes to the object, or sleep-n-touch for delays.
            # But only once, to reduce the number of API calls and the generated irrelevant events.
            # And only if the object is at least supposed to exist (not "GONE"), even if actually does not.
            if raw_event['type'] != 'DELETED':
                applied, resource_version, remaining_patch = await application.apply(
                    settings=settings,
                    resource=resource,
                    body=body,
                    patch=patch,
                    logger=local_logger,
                    delays=delays,
                    stream_pressure=stream_pressure,
                )
                if applied and matched:
                    local_logger.debug("Handling cycle is finished, waiting for new changes.")
                memory.remaining_patch = remaining_patch
                return resource_version
    return None


class _Causes(NamedTuple):
    watching_cause: causes.WatchingCause | None
    spawning_cause: causes.SpawningCause | None
    changing_cause: causes.ChangingCause | None


def _detect_causes(
        indexers: indexing.OperatorIndexers,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        resource: references.Resource,
        raw_event: bodies.RawEvent,
        body: bodies.Body,
        patch: patches.Patch,
        memory: inventory.ResourceMemory,
        local_logger: loggers.ObjectLogger,
        event_logger: loggers.ObjectLogger,
) -> _Causes:
    """Detect what are we going to do (or to skip) on this processing cycle."""

    finalizer = settings.persistence.finalizer
    extra_fields = (
        # NB: indexing handlers are useless here, they are handled on their own.
        registry._watching.get_extra_fields(resource=resource) |
        registry._changing.get_extra_fields(resource=resource) |
        registry._spawning.get_extra_fields(resource=resource))
    old = settings.persistence.diffbase_storage.fetch(body=body)
    new = settings.persistence.diffbase_storage.build(body=body, extra_fields=extra_fields)
    old = settings.persistence.progress_storage.clear(essence=old) if old is not None else None
    new = settings.persistence.progress_storage.clear(essence=new) if new is not None else None
    diff = diffs.diff(old, new)

    watching_cause = causes.detect_watching_cause(
        raw_event=raw_event,
        resource=resource,
        indices=indexers.indices,
        logger=local_logger,
        patch=patch,
        body=body,
        memo=memory.memo,
    ) if registry._watching.has_handlers(resource=resource) else None

    spawning_cause = causes.detect_spawning_cause(
        resource=resource,
        indices=indexers.indices,
        logger=event_logger,
        patch=patch,
        body=body,
        memo=memory.memo,
        reset=bool(diff),  # only essential changes reset idling, not every event
    ) if registry._spawning.has_handlers(resource=resource) else None

    changing_cause = causes.detect_changing_cause(
        finalizer=finalizer,
        raw_event=raw_event,
        resource=resource,
        indices=indexers.indices,
        logger=event_logger,
        patch=patch,
        body=body,
        old=old,
        new=new,
        diff=diff,
        memo=memory.memo,
        initial=memory.noticed_by_listing and not memory.fully_handled_once,
    ) if registry._changing.has_handlers(resource=resource) else None

    return _Causes(watching_cause, spawning_cause, changing_cause)


async def process_resource_causes(
        lifecycle: execution.LifeCycleFn,
        indexers: indexing.OperatorIndexers,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        resource: references.Resource,
        raw_event: bodies.RawEvent,
        body: bodies.Body,
        patch: patches.Patch,
        memory: inventory.ResourceMemory,
        local_logger: loggers.ObjectLogger,
        event_logger: loggers.ObjectLogger,
        stream_pressure: asyncio.Event | None,  # None for tests
        operator_paused: aiotoggles.ToggleSet | None,  # None for tests
        consistency_time: float | None,
) -> tuple[Collection[float], bool]:
    patch_initially_empty = not patch  # before we add new things in low-level handlers

    finalizer = settings.persistence.finalizer
    watching_cause, spawning_cause, changing_cause = _detect_causes(
        indexers=indexers,
        registry=registry,
        settings=settings,
        resource=resource,
        raw_event=raw_event,
        body=body,
        patch=patch,
        memory=memory,
        local_logger=local_logger,
        event_logger=event_logger,
    )

    # Invoke all the handlers that should or could be invoked at this processing cycle.
    # The low-level spies go ASAP always. However, the daemons are spawned before the high-level
    # handlers and killed after them: the daemons should live throughout the full object lifecycle.
    if watching_cause is not None:
        await process_watching_cause(
            lifecycle=lifecycles.all_at_once,
            registry=registry,
            settings=settings,
            cause=watching_cause,
        )

    spawning_delays: Collection[float] = []
    if spawning_cause is not None:
        spawning_delays = await process_spawning_cause(
            registry=registry,
            settings=settings,
            memory=memory,
            cause=spawning_cause,
            operator_paused=operator_paused,
        )

    # If there are any handlers for this resource kind in general, but not for this specific object
    # due to filters, then be blind to it, store no state, and log nothing about the handling cycle.
    if changing_cause is not None and not registry._changing.prematch(cause=changing_cause):
        changing_cause = None

    # Block the object from deletion if we have anything to do in its end of life:
    # specifically, if there are daemons to kill or mandatory on-deletion handlers to call.
    # The high-level handlers are prevented if this event cycle is dedicated to the finalizer.
    # The low-level handlers (on-event spying & daemon spawning) are still executed asap.
    deletion_is_ongoing = finalizers.is_deletion_ongoing(body=body)
    deletion_is_blocked = finalizers.is_deletion_blocked(body=body, finalizer=finalizer)
    deletion_must_be_blocked = (
        (spawning_cause is not None and
         registry._spawning.requires_finalizer(
             cause=spawning_cause,
             excluded=memory.daemons_memory.forever_stopped,
         ))
        or
        (changing_cause is not None and
         registry._changing.requires_finalizer(
             cause=changing_cause,
         )))

    if deletion_must_be_blocked and not deletion_is_blocked and not deletion_is_ongoing:
        local_logger.debug("Adding the finalizer, thus preventing the actual deletion.")
        patch.fns.append(functools.partial(finalizers.block_deletion, finalizer=finalizer))
        changing_cause = None  # prevent further high-level processing this time

    if not deletion_must_be_blocked and deletion_is_blocked:
        local_logger.debug("Removing the finalizer, as there are no handlers requiring it.")
        patch.fns.append(functools.partial(finalizers.allow_deletion, finalizer=finalizer))
        changing_cause = None  # prevent further high-level processing this time

    # If the state is inconsistent (yet), wait for new events in a hope that they bring consistency.
    # If the wait exceeds its time and no new consistent events arrive, then fake the consistency.
    # However, if a patch is accumulated by now, skip waiting and apply it instantly (by exiting).
    # In that case, we are guaranteed to be inconsistent, so also skip the state-dependent handlers.
    # Never release the object (i.e., remove the finalizer) in the inconsistent state, always wait.
    consistency_is_required = changing_cause is not None
    consistency_is_achieved = consistency_time is None  # i.e. preexisting consistency
    if changing_cause is not None and changing_cause.reason == causes.Reason.GONE:
        consistency_is_achieved = True  # for the final goodbye log message
    if consistency_is_required and not consistency_is_achieved and not patch and consistency_time:
        loop = asyncio.get_running_loop()
        unslept = await aiotime.sleep(consistency_time - loop.time(), wakeup=stream_pressure)
        consistency_is_achieved = unslept is None  # "woke up" vs. "timed out"
    consistency_is_achieved = consistency_is_achieved and patch_initially_empty
    if consistency_is_required and not consistency_is_achieved:
        return list(spawning_delays), False  # exit to PATCHing and/or re-iterating over new events.

    # Now, the consistency is either pre-proven (by receiving or not expecting any resource version)
    # or implied (by exceeding the allowed consistency-waiting timeout while getting no new events).
    # So we can go for state-dependent handlers (change detection requires a consistent state).
    changing_delays: Collection[float] = []
    if changing_cause is not None:
        changing_delays = await process_changing_cause(
            lifecycle=lifecycle,
            registry=registry,
            settings=settings,
            memory=memory,
            cause=changing_cause,
        )

    # Release the object if everything is done, and it is marked for deletion.
    # Caveat: events of type DELETED show the last finalizer as present (K8s internal logic),
    # falsely suggesting that it is still blocked and requires unblocking. No, it is not, does not.
    delays = list(spawning_delays) + list(changing_delays)
    deleted = raw_event['type'] == 'DELETED'
    if not deleted and deletion_is_ongoing and deletion_is_blocked and not delays:
        local_logger.debug("Removing the finalizer, thus allowing the actual deletion.")
        patch.fns.append(functools.partial(finalizers.allow_deletion, finalizer=finalizer))

    return (delays, changing_cause is not None)


async def process_watching_cause(
        lifecycle: execution.LifeCycleFn,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        cause: causes.WatchingCause,
) -> None:
    """
    Handle a received event, log but ignore all errors.

    This is a lightweight version of the cause handling, but for the raw events,
    without any progress persistence. Multi-step calls are also not supported.
    If the handler fails, it fails and is never retried.

    Note: K8s-event posting is skipped for ``@kopf.on.event`` handlers,
    as they should be silent. Still, the messages are logged normally.
    """
    handlers = registry._watching.get_handlers(cause=cause)
    outcomes = await execution.execute_handlers_once(
        lifecycle=lifecycle,
        settings=settings,
        handlers=handlers,
        cause=cause,
        state=progression.State.from_scratch().with_handlers(handlers),
        default_errors=execution.ErrorsMode.IGNORED,
    )

    # Store the results, but not the handlers' progress.
    progression.deliver_results(outcomes=outcomes, patch=cause.patch)


async def process_spawning_cause(
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        memory: inventory.ResourceMemory,
        cause: causes.SpawningCause,
        operator_paused: aiotoggles.ToggleSet | None,  # None for tests
) -> Collection[float]:
    """
    Spawn/kill all the background tasks of a resource.

    The spawning and killing happens in parallel with the resource-changing
    handlers invocation (even if it takes a few cycles). For this, the signal
    to terminate is sent to the daemons immediately, but the actual check
    of their shutdown is performed only when all the on-deletion handlers
    have succeeded (or after they were invoked if they are optional;
    or immediately if there were no on-deletion handlers to invoke at all).

    The resource remains blocked by the finalizers until all the daemons exit
    (except those marked as tolerating being orphaned).
    """

    # Refresh the up-to-date body & essential timestamp for all the daemons/timers.
    if memory.daemons_memory.live_fresh_body is None:
        memory.daemons_memory.live_fresh_body = cause.body
    if cause.reset:
        memory.daemons_memory.idle_reset_time = asyncio.get_running_loop().time()

    if finalizers.is_deletion_ongoing(cause.body):
        stopping_delays = await daemons.stop_daemons(
            settings=settings,
            daemons=memory.daemons_memory.running_daemons,
        )
        return stopping_delays

    else:
        handlers = registry._spawning.get_handlers(
            cause=cause,
            excluded=memory.daemons_memory.forever_stopped,
        )
        spawning_delays = await daemons.spawn_daemons(
            settings=settings,
            daemons=memory.daemons_memory.running_daemons,
            cause=cause,
            memory=memory.daemons_memory,
            handlers=handlers,
        )
        matching_delays = await daemons.match_daemons(
            settings=settings,
            daemons=memory.daemons_memory.running_daemons,
            handlers=handlers,
        )
        # Critical: strictly after spawning; see the docstring why.
        pausing_delays = await daemons.pause_daemons(
            settings=settings,
            daemons=memory.daemons_memory.running_daemons,
            operator_paused=operator_paused,
        )
        return list(spawning_delays) + list(matching_delays) + list(pausing_delays)


async def process_changing_cause(
        lifecycle: execution.LifeCycleFn,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        memory: inventory.ResourceMemory,
        cause: causes.ChangingCause,
) -> Collection[float]:
    """
    Handle a detected cause as part of the broader handler routine.
    """
    logger = cause.logger
    patch = cause.patch  # TODO get rid of this alias
    body = cause.body  # TODO get rid of this alias
    delays: Collection[float] = []
    done: bool | None = None
    skip: bool | None = None

    # Regular causes invoke the handlers.
    if cause.reason in causes.HANDLER_REASONS:
        title = causes.TITLES.get(cause.reason.value, repr(cause.reason.value))

        resource_registry = registry._changing
        owned_handlers = resource_registry.get_resource_handlers(resource=cause.resource)
        cause_handlers = resource_registry.get_handlers(cause=cause)
        storage = settings.persistence.progress_storage
        state = progression.State.from_storage(body=cause.body, storage=storage, handlers=owned_handlers)
        state = state.with_purpose(cause.reason).with_handlers(cause_handlers)

        # Report the causes that have been superseded (intercepted, overridden) by the current one.
        # The mix-in causes (i.e. resuming) is re-purposed if its handlers are still selected.
        # To the next cycle, all extras are purged or re-purposed, so the message does not repeat.
        for extra_purpose, counters in state.extras.items():  # usually 0..1 items, rarely 2+.
            extra_title = causes.TITLES.get(extra_purpose, repr(extra_purpose))
            logger.info(f"{extra_title.capitalize()} is superseded by {title.lower()}: "
                        f"{counters.success} succeeded; "
                        f"{counters.failure} failed; "
                        f"{counters.running} left to the moment.")
            state = state.with_purpose(purpose=cause.reason, handlers=cause_handlers)

        # Purge the now-irrelevant handlers if they were not re-purposed (extras are recalculated!).
        # The current cause continues afterwards, and overrides its own pre-purged handler states.
        # TODO: purge only the handlers that fell out of current purpose; but it is not critical
        if state.extras:
            state.purge(body=cause.body, patch=cause.patch,
                        storage=storage, handlers=owned_handlers)

        # Inform on the current cause/event on every processing cycle. Even if there are
        # no handlers -- to show what has happened and why the diff-base is patched.
        logger.debug(f"{title.capitalize()} is in progress: {body!r}")
        if cause.diff and cause.old is not None and cause.new is not None:
            logger.debug(f"{title.capitalize()} diff: {cause.diff!r}")

        if cause_handlers:
            outcomes = await execution.execute_handlers_once(
                lifecycle=lifecycle,
                settings=settings,
                handlers=cause_handlers,
                cause=cause,
                state=state,
                extra_context=subhandling.subhandling_context,
            )
            state = state.with_outcomes(outcomes)
            state.store(body=cause.body, patch=cause.patch, storage=storage)
            progression.deliver_results(outcomes=outcomes, patch=cause.patch)

            if state.done:
                counters = state.counts  # calculate only once
                logger.info(f"{title.capitalize()} is processed: "
                            f"{counters.success} succeeded; "
                            f"{counters.failure} failed.")
                state.purge(body=cause.body, patch=cause.patch,
                            storage=storage, handlers=owned_handlers)

            done = state.done
            delays = state.delays
        else:
            skip = True

    # Regular causes also do some implicit post-handling when all handlers are done.
    if done or skip:
        if cause.new is not None and cause.old != cause.new:
            settings.persistence.diffbase_storage.store(body=body, patch=patch, essence=cause.new)

        # Once all handlers have succeeded at least once for any reason, or if there were none,
        # prevent further resume-handlers (which otherwise happens on each watch-stream re-listing).
        memory.fully_handled_once = True

    # Informational causes just print the log lines.
    if cause.reason == causes.Reason.GONE:
        logger.debug("Deleted, really deleted, and we are notified.")

    if cause.reason == causes.Reason.FREE:
        logger.debug("Deletion, but we are done with it, and we do not care.")

    if cause.reason == causes.Reason.NOOP:
        logger.debug("Something has changed, but we are not interested (the essence is the same).")

    # The delay is then consumed by the main handling routine (in different ways).
    return delays



"""
All the functions to manipulate the resource fields, state changes, etc.

Grouped by the type of the fields and the purpose of the manipulation.

Used in the handling routines to check if there were significant changes at all
(i.e. not our own internal and system changes, like the uids, links, etc),
and to get the exact per-field diffs for the specific handler functions.

All the functions are purely data-manipulative and computational.
No external calls or any I/O activities are done here.
"""



import asyncio
import gc
import warnings
from unittest.mock import AsyncMock, Mock

import pytest

from kopf._cogs.aiokits.aiotasks import cancel_coro


async def f(mock):
    return mock()


# Kwargs are accepted to match the signatures, but are unused/not passed through due to no need.
# Usually those are `name` & `context`, as in `asyncio.create_task(…)`.
def factory(loop, coro_or_mock, **_):
    coro = coro_or_mock._mock_wraps if isinstance(coro_or_mock, AsyncMock) else coro_or_mock
    return asyncio.Task(coro, loop=loop)


@pytest.fixture(autouse=True)
async def coromock_task_factory():
    factory_spy = Mock(wraps=factory)
    asyncio.get_running_loop().set_task_factory(factory_spy)
    yield factory_spy
    asyncio.get_running_loop().set_task_factory(None)


async def test_coro_issues_a_warning_normally(coromock_task_factory):
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('default')
        mock = Mock()
        coro = f(mock)

        # The warnings come only from the garbage collection, so dereference it.
        del coro
        gc.collect()

    # The 1st coro is the test function itself; the 2nd coro would be the coro-under-test.
    assert coromock_task_factory.call_count == 1  # i.e. the task was NOT created.
    assert not mock.called
    assert len(w) == 1
    assert issubclass(w[0].category, RuntimeWarning)
    assert str(w[0].message) == "coroutine 'f' was never awaited"


async def test_coro_is_closed_via_a_hack_with_no_warning(coromock_task_factory):
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('default')
        mock = Mock()
        coro = f(mock)
        await cancel_coro(coro)

        # The warnings come only from the garbage collection, so dereference it.
        del coro
        gc.collect()

    # The 1st coro is the test function itself; the 2nd coro would be the coro-under-test.
    assert coromock_task_factory.call_count == 1  # i.e. the task was NOT created.
    assert not mock.called
    assert not w


async def test_coro_is_awaited_via_a_task_with_no_warning(coromock_task_factory):
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('default')
        mock = Mock()
        coro = f(mock)
        coromock = AsyncMock(wraps=coro, spec=coro)
        del coromock.close
        await cancel_coro(coromock)

        # The warnings come only from the garbage collection, so dereference it.
        del coro
        gc.collect()

    # The 1st coro is the test function itself; the 2nd coro is the coro-under-test.
    assert coromock_task_factory.call_count == 2  # i.e. the task WAS created.
    assert not mock.called
    assert not w



"""
A few simple lifecycles for the handlers.

New lifecycles can be implemented the same way: accept ``handlers``
in the order they are registered (except those already succeeded),
and return the list of handlers in the order and amount to be executed.

The default behaviour of the framework is the most simplistic:
execute in the order they are registered, one by one.
"""
import logging
import random
from collections.abc import Sequence
from typing import Any

from kopf._core.actions import execution

logger = logging.getLogger(__name__)

Handlers = Sequence[execution.Handler]


def all_at_once(handlers: Handlers, **_: Any) -> Handlers:
    """ Execute all handlers at once, in one event reaction cycle, if possible. """
    return handlers


def one_by_one(handlers: Handlers, **_: Any) -> Handlers:
    """ Execute handlers one at a time, in the order they were registered. """
    return handlers[:1]


def randomized(handlers: Handlers, **_: Any) -> Handlers:
    """ Execute one handler at a time, in the random order. """
    return [random.choice(handlers)] if handlers else []


def shuffled(handlers: Handlers, **_: Any) -> Handlers:
    """ Execute all handlers at once, but in the random order. """
    return random.sample(handlers, k=len(handlers)) if handlers else []


def asap(handlers: Handlers, *, state: execution.State, **_: Any) -> Handlers:
    """ Execute one handler at a time, skip on failure, try the next one, retry after the full cycle. """

    def keyfn(handler: execution.Handler) -> int:
        return state[handler.id].retries or 0

    return sorted(handlers, key=keyfn)[:1]


_default_lifecycle: execution.LifeCycleFn = asap


def get_default_lifecycle() -> execution.LifeCycleFn:
    return _default_lifecycle


def set_default_lifecycle(lifecycle: execution.LifeCycleFn | None) -> None:
    global _default_lifecycle
    if _default_lifecycle is not None:
        logger.warning(f"The default lifecycle is already set to {_default_lifecycle}, overriding it to {lifecycle}.")
    _default_lifecycle = lifecycle if lifecycle is not None else asap



import asyncio

import kopf


async def test_timer_is_spawned_at_least_once(
        resource, dummy, assert_logs, k8s_mocked, simulate_cycle, looptime):
    trigger = asyncio.Condition()

    @kopf.timer(*resource, id='fn', interval=1.0)
    async def fn(**kwargs):
        dummy.mock(**kwargs)
        async with trigger:
            trigger.notify_all()

    await simulate_cycle({})
    async with trigger:
        await trigger.wait()
        await trigger.wait()

    assert looptime == 1
    assert dummy.mock.call_count == 2


async def test_timer_initial_delay_obeyed(
        resource, dummy, assert_logs, k8s_mocked, simulate_cycle, looptime):
    trigger = asyncio.Condition()

    @kopf.timer(*resource, id='fn', initial_delay=5.0, interval=1.0)
    async def fn(**kwargs):
        dummy.mock(**kwargs)
        async with trigger:
            trigger.notify_all()

    await simulate_cycle({})
    async with trigger:
        await trigger.wait()
        await trigger.wait()

    assert looptime == 6
    assert dummy.mock.call_count == 2


async def test_timer_initial_delay_callable_obeyed(
        resource, dummy, assert_logs, k8s_mocked, simulate_cycle, looptime):
    trigger = asyncio.Condition()

    def get_delay(body, **_):
        return body.get('spec', {}).get('delay', 0.0)

    @kopf.timer(*resource, id='fn', initial_delay=get_delay, interval=1.0)
    async def fn(**kwargs):
        dummy.mock(**kwargs)
        async with trigger:
            trigger.notify_all()

    await simulate_cycle({'spec': {'delay': 3.0}})
    async with trigger:
        await trigger.wait()
        await trigger.wait()

    assert looptime == 4.0 # 3.0 delay + 1.0 interval
    assert dummy.mock.call_count == 2



import asyncio

import aiohttp.web
import pytest

from kopf._cogs.clients.api import request
from kopf._cogs.clients.errors import APIError

pytestmark = pytest.mark.usefixtures('fake_vault')


@pytest.fixture(autouse=True)
def sleep(mocker):
    """Do not let it actually sleep, even if it is a 0-sleep."""
    return mocker.patch('asyncio.sleep')


@pytest.fixture()
def request_fn(mocker):
    return mocker.patch('aiohttp.ClientSession.request')


async def test_regular_errors_escalate_without_retries(
        assert_logs, settings, logger, request_fn):
    request_fn.side_effect = Exception("boo")

    settings.networking.error_backoffs = [1, 2, 3]
    with pytest.raises(Exception) as err:
        await request('get', '/url', settings=settings, logger=logger)

    assert str(err.value) == "boo"
    assert request_fn.call_count == 1
    assert_logs(prohibited=["attempt", "escalating", "retry"])


# All client errors except explicitly retryable (403, 429).
@pytest.mark.parametrize('status', [400, 404, 409, 499, 666, 999])
async def test_permanent_errors_escalate_without_retries(
        assert_logs, settings, logger, kmock, status):
    kmock['get /url'] << {} << status

    settings.networking.error_backoffs = [1, 2, 3]
    with pytest.raises(APIError) as err:
        await request('get', '/url', settings=settings, logger=logger)

    assert err.value.status == status
    assert len(kmock) == 1
    assert_logs(prohibited=["attempt", "escalating", "retry"])


# All server errors, certain client errors (403, 429).
@pytest.mark.parametrize('status', [403, 429, 500, 503, 599])
async def test_temporary_errors_escalate_with_retries(
        assert_logs, settings, logger, kmock, status):
    kmock['get /url'] << {} << status

    settings.networking.error_backoffs = [0, 0, 0]
    with pytest.raises(APIError) as err:
        await request('get', '/url', settings=settings, logger=logger)

    assert err.value.status == status
    assert len(kmock) == 4
    assert_logs([
        "attempt #1/4 failed; will retry",
        "attempt #2/4 failed; will retry",
        "attempt #3/4 failed; will retry",
        "attempt #4/4 failed; escalating",
    ])


async def test_connection_errors_escalate_with_retries(assert_logs, settings, logger, request_fn):
    request_fn.side_effect = aiohttp.ClientConnectionError()

    settings.networking.error_backoffs = [0, 0, 0]
    with pytest.raises(aiohttp.ClientConnectionError):
        await request('get', '/url', settings=settings, logger=logger)

    assert request_fn.call_count == 4
    assert_logs([
        "attempt #1/4 failed; will retry",
        "attempt #2/4 failed; will retry",
        "attempt #3/4 failed; will retry",
        "attempt #4/4 failed; escalating",
    ])


async def test_timeout_errors_escalate_with_retries(assert_logs, settings, logger, request_fn):
    request_fn.side_effect = asyncio.TimeoutError()

    settings.networking.error_backoffs = [0, 0, 0]
    with pytest.raises(asyncio.TimeoutError):
        await request('get', '/url', settings=settings, logger=logger)

    assert request_fn.call_count == 4
    assert_logs([
        "attempt #1/4 failed; will retry",
        "attempt #2/4 failed; will retry",
        "attempt #3/4 failed; will retry",
        "attempt #4/4 failed; escalating",
    ])


async def test_retried_until_succeeded(assert_logs, settings, logger, kmock):
    api1 = kmock['get /url'][:2] << {} << 505
    api2 = kmock['get /url'] << {}

    settings.networking.error_backoffs = [0, 0, 0]
    await request('get', '/url', settings=settings, logger=logger)

    assert len(api1) == 2  # 2 failures, 1 success
    assert len(api2) == 1  # 1 success; the other one is not requested
    assert_logs([
        "attempt #1/4 failed; will retry",
        "attempt #2/4 failed; will retry",
        "attempt #3/4 succeeded",
    ], prohibited=[
        "attempt #4/4",
    ])


@pytest.mark.parametrize('backoffs, exp_calls', [
    ([], 1),
    ([0], 2),
    ([1], 2),
    ([0, 0], 3),
    ([1, 2], 3),
])
@pytest.mark.parametrize('status', [429, 500])
async def test_backoffs_as_lists(settings, logger, kmock, sleep, backoffs, exp_calls, status):
    kmock['get /url'][:4] << {} << status

    settings.networking.error_backoffs = backoffs
    with pytest.raises(APIError):
        await request('get', '/url', settings=settings, logger=logger)

    assert len(kmock) == exp_calls
    all_sleeps = [call.args[0] for call in sleep.call_args_list]
    assert all_sleeps == backoffs


async def test_backoffs_as_floats(settings, logger, kmock, sleep):
    kmock['get /url'] << {} << 500

    settings.networking.error_backoffs = 5.0
    with pytest.raises(APIError):
        await request('get', '/url', settings=settings, logger=logger)

    assert len(kmock) == 2
    all_sleeps = [call.args[0] for call in sleep.call_args_list]
    assert all_sleeps == [5.0]


async def test_backoffs_as_iterables(settings, logger, kmock, sleep):
    kmock['get /url'][:8] << {} << 500

    class Itr:
        def __iter__(self):
            return iter([1, 2, 3])

    settings.networking.error_backoffs = Itr()  # to be reused on every attempt
    with pytest.raises(APIError):
        await request('get', '/url', settings=settings, logger=logger)
    with pytest.raises(APIError):
        await request('get', '/url', settings=settings, logger=logger)

    assert len(kmock) == 8
    all_sleeps = [call.args[0] for call in sleep.call_args_list]
    assert all_sleeps == [1, 2, 3, 1, 2, 3]



@pytest.mark.parametrize('headers, payload', [
    pytest.param({'Retry-After': '10'}, {}, id='new-style'),
    pytest.param({}, {'kind': 'Status', 'details': {'retryAfterSeconds': 10}}, id='old-style'),
])
async def test_retry_after_overrides_backoffs_when_enforced(
        assert_logs, settings, logger, kmock, sleep, headers, payload):
    kmock['get /url'] << 429 << kmock.headers(headers) << payload

    settings.networking.enforce_retry_after = True
    settings.networking.error_backoffs = [5, 10, 15]
    with pytest.raises(APIError):
        await request('get', '/url', settings=settings, logger=logger)

    assert_logs([
        "Overriding the backoff 5s with the retry-after 10s",
        "attempt #1/4 failed; will retry",
        "Overriding the backoff 10s with the retry-after 10s",
        "attempt #2/4 failed; will retry",
        "Overriding the backoff 15s with the retry-after 10s",
        "attempt #3/4 failed",
        "attempt #4/4 failed; escalating",
    ])

    assert len(kmock) == 4
    all_sleeps = [call.args[0] for call in sleep.call_args_list]
    assert all_sleeps == [10, 10, 10]  # enforced retry-after


@pytest.mark.parametrize('headers, payload', [
    pytest.param({'Retry-After': '10'}, {}, id='new-style'),
    pytest.param({}, {'kind': 'Status', 'details': {'retryAfterSeconds': 10}}, id='old-style'),
])
async def test_retry_after_overrides_backoffs_when_longer(caplog,
        assert_logs, settings, logger, kmock, sleep, headers, payload):
    kmock['get /url'] << 429 << kmock.headers(headers) << payload

    settings.networking.error_backoffs = [5, 10, 15]
    with pytest.raises(APIError):
        await request('get', '/url', settings=settings, logger=logger)

    assert_logs([
        "Overriding the backoff 5s with the retry-after 10s",  # because retry-after is bigger
        "attempt #1/4 failed; will retry",
        "attempt #2/4 failed; will retry",
        "attempt #3/4 failed",
        "attempt #4/4 failed",
    ], prohibited=[
        "Overriding the backoff 10s with the retry-after 10s",
        "Overriding the backoff 15s with the retry-after 10s",
    ])

    assert len(kmock) == 4
    all_sleeps = [call.args[0] for call in sleep.call_args_list]
    assert all_sleeps == [10, 10, 15]  # never smaller than retry-after



from kopf._cogs.clients import api, errors
from kopf._cogs.configs import configuration
from kopf._cogs.helpers import typedefs
from kopf._cogs.structs import bodies, patches, references


async def patch_obj(
        *,
        settings: configuration.OperatorSettings,
        resource: references.Resource,
        namespace: references.Namespace,
        name: str | None,
        patch: patches.Patch,
        logger: typedefs.Logger,
        silent: bool = False,
) -> tuple[bodies.RawBody | None, patches.Patch | None]:
    """
    Patch a resource of specific kind.

    Unlike the object listing, the namespaced call is always
    used for the namespaced resources, even if the operator serves
    the whole cluster (i.e. is not namespace-restricted).

    Returns the patched body. The patched body can be partial (status-only,
    no-status, or empty) --- depending on whether there were fields in the body
    or in the status to patch; if neither had fields for patching, the result
    is an empty body. The result should only be used to check against the patch:
    if there was nothing to patch, it does not matter if the fields are absent.

    Returns ``None`` if the underlying object is absent, as detected by trying
    to patch it and failing with HTTP 404. This can happen if the object was
    deleted in the operator's handlers or externally during the processing,
    so that the framework was unaware of these changes until the last moment.
    """
    as_subresource = 'status' in resource.subresources
    body_patch = dict(patch)  # shallow: for mutation of the top-level keys below.
    status_patch = body_patch.pop('status', None) if as_subresource else None
    status_patch = {'status': status_patch} if status_patch is not None else None

    # Patch & reconstruct the actual body as reported by the server. The reconstructed body can be
    # partial or empty -- if the body/status patches are empty. This is fine: it is only used
    # to verify that the patched fields are matching the patch. No patch? No mismatch!
    try:
        patched_body: bodies.RawBody | None = None

        if body_patch:
            if not silent:
                logger.debug(f"Merge-patching the resource with: {body_patch!r}")
            patched_body = await api.patch(
                url=resource.get_url(namespace=namespace, name=name),
                headers={'Content-Type': 'application/merge-patch+json'},
                payload=body_patch,
                settings=settings,
                logger=logger,
            )

        if status_patch:
            # NB: we need the new resourceVersion, so we take the whole new patched body.
            if not silent:
                logger.debug(f"Merge-patching the status with: {status_patch!r}")
            patched_body = await api.patch(
                url=resource.get_url(namespace=namespace, name=name, subresource='status'),
                headers={'Content-Type': 'application/merge-patch+json'},
                payload=status_patch,
                settings=settings,
                logger=logger,
            )

        # Only the callable transformations are left now, no dict-merge components left.
        # If we fail at any stage below, we re-apply this whole patch. We cannot distinguish
        # the "status" functions from the "main body" functions; some of them can be mixed.
        remaining_patch = patches.Patch(fns=patch.fns)

        # Calculate the JSON diffs to be applied and split into body + status (if a subresource).
        # Use the latest known body for reference when calculating the item indexes in lists.
        # Note: we apply transformations for the whole body atomically, and in the future (not yet)
        # we want to get the ops from the functions, so we can only filter by path here.
        fresh_body: bodies.RawBody | None = patched_body if patched_body else patch._original
        ops: patches.JSONPatch = remaining_patch.as_json_patch(fresh_body)
        body_ops: patches.JSONPatch
        status_ops: patches.JSONPatch
        if as_subresource:
            body_ops = [op for op in ops
                        if not(op['path'] == '/status' or op['path'].startswith('/status/'))]
            status_ops = [op for op in ops
                          if op['path'] == '/status' or op['path'].startswith('/status/')]
        else:
            body_ops = ops
            status_ops = []

        # Apply pending operations (finalizers) via JSON-patch with optimistic concurrency.
        # If the resource version fails the test, we know there are newer changes pending,
        # so we will get back to the handling cycle & this same patching almost immediately.
        # NB: the patched_body=={} in tests' mocks, so check for emptiness, not just null-ness.
        if body_ops:
            resource_version = (fresh_body or {}).get('metadata', {}).get('resourceVersion')
            test = patches.JSONPatchItem(op='test', path='/metadata/resourceVersion', value=resource_version)
            ops = [test] + body_ops
            if not silent:
                logger.debug(f"JSON-patching the resource with: {ops!r}")
            try:
                patched_body = await api.patch(
                    url=resource.get_url(namespace=namespace, name=name),
                    headers={'Content-Type': 'application/json-patch+json'},
                    payload=ops,
                    settings=settings,
                    logger=logger,
                )
            except errors.APIUnprocessableEntityError:
                # NB: also detach from the current freshest body, persist only the patch fns.
                if not silent:
                    logger.debug(
                        "Could not apply the patch in full due to conflicts with newer changes. "
                        f"Will try on the next cycle soon. Remaining: {remaining_patch!r}"
                    )
                return patched_body, remaining_patch
            else:
                fresh_body = patched_body

        # We DO NOT recalculate the diff against a newer body after the recent patch!
        # The ops of both diffs are already non-conflicting, because we have split them so.
        # If the body patch succeeds, we know the new fresh state. If it fails, we do not get here.
        # If something jumps inbetween, we fail & postpone the whole JSON-patch to the next cycle.
        if status_ops:
            resource_version = (fresh_body or {}).get('metadata', {}).get('resourceVersion')
            test = patches.JSONPatchItem(op='test', path='/metadata/resourceVersion', value=resource_version)
            ops = [test] + status_ops
            if not silent:
                logger.debug(f"JSON-patching the status with: {ops!r}")
            try:
                patched_body = await api.patch(
                    url=resource.get_url(namespace=namespace, name=name, subresource='status'),
                    headers={'Content-Type': 'application/json-patch+json'},
                    payload=ops,
                    settings=settings,
                    logger=logger,
                )
            except errors.APIUnprocessableEntityError:
                # NB: also detach from the current freshest body, persist only the patch fns.
                if not silent:
                    logger.debug(
                        "Could not apply the patch in full due to conflicts with newer changes. "
                        f"Will try on the next cycle soon. Remaining: {remaining_patch!r}"
                    )
                return patched_body, remaining_patch

        return patched_body, None

    except errors.APINotFoundError:
        logger.debug(f"Patching was skipped: the object does not exist anymore.")
        return None, None



import enum
from collections.abc import Mapping
from typing import Any, TypeAlias

from kopf._core.intents import callbacks


class MetaFilterToken(enum.Enum):
    """ Tokens for filtering by annotations/labels. """
    PRESENT = enum.auto()
    ABSENT = enum.auto()


# For exporting to the top-level package.
ABSENT = MetaFilterToken.ABSENT
PRESENT = MetaFilterToken.PRESENT

# Filters for handler specifications (not the same as the object's values).
MetaFilter = Mapping[str, str | MetaFilterToken | callbacks.MetaFilterFn]

# Filters for old/new values of a field.
# NB: `Any` covers all other values, but we want to highlight that they are specially treated.
ValueFilter: TypeAlias = Any | MetaFilterToken | callbacks.MetaFilterFn | None



"""
All the structures coming from/to the Kubernetes API.

The usage of these classes is spread over the codebase, so they are extracted
into a separate module of such type definitions.

For strict type-checking, they are detailed to the per-field level
(e.g. ``TypedDict`` instead of just ``Mapping[Any, Any]``) --
as used by the framework. The operators can use arbitrary fields at runtime,
which are not declared in the type definitions at type-checking time.

In case the operators are also type-checked, type casting can be used
(without ``cast()``, this code fails at type-checking, though works at runtime):

.. code-block:: python

    from typing import cast
    import kopf

    class MyMeta(kopf.Meta):
        unknownField: str

    @kopf.on.create('kopfexamples')
    def create_fn(*args, meta: kopf.Meta, **kwargs):
        meta = cast(MyMeta, meta)
        print(meta['unknownField'])

.. note::

    There is a strict separation of objects coming from/to the Kubernetes API
    and from (but not to) the users:

    The Kubernetes-originated objects are dicts or dict-like custom classes.
    The framework internally expects them to be such. Arbitrary 3rd-party
    classes are not supported and are not delivered to the handlers.

    The user-originated objects can be either one of the Kubernetes-originated
    framework-supported types (dicts/dict-like), or a 3rd-party class,
    such as from ``pykube-ng``, ``kubernetes`` client, etc --- as long as it is
    supported by the framework's object-processing functions.

    In the future, extra classes can be added for the user-originated objects
    and object-processing functions. The internal dicts will remain the same.
"""

from collections.abc import Mapping
from typing import Any, Literal, TypeAlias, TypedDict, cast

from kopf._cogs.structs import dicts, references

# Make sure every kwarg has a corresponding same-named type in the root package.
# Usually, they are "live views" into dicts, so they are not dicts themselves.
Labels: TypeAlias = Mapping[str, str]
Annotations: TypeAlias = Mapping[str, str]

#
# Everything marked "raw" is a plain unwrapped unprocessed data as JSON-decoded
# from Kubernetes API, usually as retrieved in watching or fetching API calls.
# "Input" is a parsed JSON as is, while "event" is an "input" without "errors".
# All non-used payload falls into `Any`, and is not type-checked.
#

# ``None`` is used for the listing, when the pseudo-watch-stream is simulated.
RawInputType = Literal[None, 'ADDED', 'MODIFIED', 'DELETED', 'BOOKMARK', 'ERROR']
RawEventType = Literal[None, 'ADDED', 'MODIFIED', 'DELETED', 'BOOKMARK']


class RawMeta(TypedDict, total=False):
    uid: str
    name: str
    namespace: str
    labels: dict[str, str]
    annotations: dict[str, str]
    finalizers: list[str]
    resourceVersion: str
    deletionTimestamp: str
    creationTimestamp: str
    selfLink: str


class RawBody(TypedDict, total=False):
    apiVersion: str
    kind: str
    metadata: RawMeta
    spec: dict[str, Any]
    status: dict[str, Any]


# A special payload for type==ERROR (this is not a connection or client error).
class RawError(TypedDict, total=False):
    apiVersion: str     # usually: Literal['v1']
    kind: str           # usually: Literal['Status']
    metadata: dict[Any, Any]
    code: int
    reason: str
    status: str
    message: str


# As received from the stream before processing the errors and special cases.
class RawInput(TypedDict, total=True):
    type: RawInputType
    object: RawBody | RawError


# As passed to the framework after processing the errors and special cases.
class RawEvent(TypedDict, total=True):
    type: RawEventType
    object: RawBody


#
# Body/Meta essences only contain the fields relevant for object diff tracking.
# They are presented to the user as part of the diff's `old`/`new` fields & kwargs.
# Added for stricter type checking, to differentiate from the actual Body/Meta.
#


class MetaEssence(TypedDict, total=False):
    labels: dict[str, str]
    annotations: dict[str, str]


class BodyEssence(TypedDict, total=False):
    metadata: MetaEssence
    spec: dict[str, Any]
    status: dict[str, Any]


#
# Enhanced dict-wrappers for easier typed access to well-known typed fields,
# with live view of updates and changes in the root body (for daemon's)
# Despite they are just MappingViews with no extensions, they are separated
# for stricter typing of arguments.
#


class Meta(dicts.MappingView[str, Any]):

    def __init__(self, __src: "Body") -> None:
        super().__init__(__src, 'metadata')
        self._labels: dicts.MappingView[str, str] = dicts.MappingView(self, 'labels')
        self._annotations: dicts.MappingView[str, str] = dicts.MappingView(self, 'annotations')

    @property
    def labels(self) -> Labels:
        return self._labels

    @property
    def annotations(self) -> Annotations:
        return self._annotations

    @property
    def uid(self) -> str | None:
        return cast(str | None, self.get('uid'))

    @property
    def name(self) -> str | None:
        return cast(str | None, self.get('name'))

    @property
    def namespace(self) -> references.Namespace:
        return cast(references.Namespace, self.get('namespace'))

    @property
    def creation_timestamp(self) -> str | None:
        return cast(str | None, self.get('creationTimestamp'))

    @property
    def deletion_timestamp(self) -> str | None:
        return cast(str | None, self.get('deletionTimestamp'))


class Spec(dicts.MappingView[str, Any]):
    def __init__(self, __src: "Body") -> None:
        super().__init__(__src, 'spec')


class Status(dicts.MappingView[str, Any]):
    def __init__(self, __src: "Body") -> None:
        super().__init__(__src, 'status')


class Body(dicts.ReplaceableMappingView[str, Any]):

    def __init__(self, __src: RawBody | BodyEssence) -> None:
        super().__init__(__src)
        self._meta = Meta(self)
        self._spec = Spec(self)
        self._status = Status(self)

    @property
    def metadata(self) -> Meta:
        return self._meta

    @property
    def meta(self) -> Meta:
        return self._meta

    @property
    def spec(self) -> Spec:
        return self._spec

    @property
    def status(self) -> Status:
        return self._status


#
# Other API types, which are not body parts.
#

class ObjectReference(TypedDict, total=False):
    apiVersion: str
    kind: str
    namespace: str | None
    name: str
    uid: str


class OwnerReference(TypedDict, total=False):
    controller: bool
    blockOwnerDeletion: bool
    apiVersion: str
    kind: str
    name: str
    uid: str


def build_object_reference(
        body: Body,
) -> ObjectReference:
    """
    Construct an object reference for the events.

    Keep in mind that some fields can be absent: e.g. ``namespace``
    for cluster resources, or e.g. ``apiVersion`` for ``kind: Node``, etc.
    """
    ref = dict(
        apiVersion=body.get('apiVersion'),
        kind=body.get('kind'),
        name=body.get('metadata', {}).get('name'),
        uid=body.get('metadata', {}).get('uid'),
        namespace=body.get('metadata', {}).get('namespace'),
    )
    return cast(ObjectReference, {key: val for key, val in ref.items() if val})


def build_owner_reference(
        body: Body,
        *,
        controller: bool | None = True,
        block_owner_deletion: bool | None = True,
) -> OwnerReference:
    """
    Construct an owner reference object for the parent-children relationships.

    The structure needed to link the children objects to the current object as a parent.
    See https://kubernetes.io/docs/concepts/workloads/controllers/garbage-collection/

    Keep in mind that some fields can be absent: e.g. ``namespace``
    for cluster resources, or e.g. ``apiVersion`` for ``kind: Node``, etc.
    """
    ref = dict(
        controller=controller,
        blockOwnerDeletion=block_owner_deletion,
        apiVersion=body.get('apiVersion'),
        kind=body.get('kind'),
        name=body.get('metadata', {}).get('name'),
        uid=body.get('metadata', {}).get('uid'),
    )
    return cast(OwnerReference, {key: val for key, val in ref.items() if val is not None})



"""
All the routines to talk to Kubernetes API and other APIs.

This library is supposed to be mocked when the mocked K8s client is needed,
and only the high-level logic has to be tested, not the API calls themselves.

Beware: this is NOT a Kubernetes client. It is a set of dedicated adapters
specially tailored to do the framework-specific tasks, not the generic
Kubernetes object manipulation.

The operators MUST NOT rely on how the framework communicates with the cluster.
Specifically:

Currently, all the routines use the official Kubernetes client.
Eventually, it can be replaced with anything else (e.g. pykube-ng).

Currently, most of the routines are synchronous, i.e. blocking
from the asyncio's point of view. Later, they can be replaced
to async coroutines (if the client supports that),
or put into the asyncio executors (thread pools).
"""



import asyncio

import freezegun
import pytest

from kopf._cogs.aiokits import aiotoggles
from kopf._cogs.structs import bodies
from kopf._core.engines.peering import process_peering_event


async def test_other_peering_objects_are_ignored(
        mocker, k8s_mocked, assert_logs, settings, looptime,
        peering_resource, peering_namespace):

    status = mocker.Mock()
    status.items.side_effect = Exception("This should not be called.")
    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'their-name'},
            'status': status,
        })

    settings.peering.name = 'our-name'
    await process_peering_event(
        raw_event=event,
        autoclean=False,
        identity='id',
        settings=settings,
        namespace=peering_namespace,
        resource=peering_resource,
    )
    assert not status.items.called
    assert not k8s_mocked.patch.called
    assert looptime == 0
    assert_logs(prohibited=["patching"])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_toggled_on_for_higher_priority_peer_when_initially_off(
        k8s_mocked, assert_logs, settings, looptime,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 101,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    conflicts_found = aiotoggles.Toggle(False)
    stream_pressure = asyncio.Event()
    loop = asyncio.get_running_loop()
    loop.call_later(1.23, stream_pressure.set)

    assert conflicts_found.is_off()
    await process_peering_event(
        raw_event=event,
        conflicts_found=conflicts_found,
        stream_pressure=stream_pressure,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert conflicts_found.is_on()
    assert looptime == 1.23
    assert not k8s_mocked.patch.called
    assert_logs(["Pausing operations in favour of"], prohibited=[
        "Possibly conflicting operators",
        "Pausing all operators, including self",
        "Resuming operations after the pause",
        "patching",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_ignored_for_higher_priority_peer_when_already_on(
        k8s_mocked, assert_logs, settings, looptime,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 101,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    conflicts_found = aiotoggles.Toggle(True)
    stream_pressure = asyncio.Event()
    loop = asyncio.get_running_loop()
    loop.call_later(1.23, stream_pressure.set)

    assert conflicts_found.is_on()
    await process_peering_event(
        raw_event=event,
        conflicts_found=conflicts_found,
        stream_pressure=stream_pressure,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert conflicts_found.is_on()
    assert looptime == 1.23
    assert not k8s_mocked.patch.called
    assert_logs(prohibited=[
        "Possibly conflicting operators",
        "Pausing all operators, including self",
        "Pausing operations in favour of",
        "Resuming operations after the pause",
        "patching",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_toggled_off_for_lower_priority_peer_when_initially_on(
        k8s_mocked, assert_logs, settings, looptime,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 99,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    conflicts_found = aiotoggles.Toggle(True)
    stream_pressure = asyncio.Event()
    loop = asyncio.get_running_loop()
    loop.call_later(1.23, stream_pressure.set)

    assert conflicts_found.is_on()
    await process_peering_event(
        raw_event=event,
        conflicts_found=conflicts_found,
        stream_pressure=stream_pressure,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert conflicts_found.is_off()
    assert looptime == 0
    assert not k8s_mocked.patch.called
    assert_logs(["Resuming operations after the pause"], prohibited=[
        "Possibly conflicting operators",
        "Pausing all operators, including self",
        "Pausing operations in favour of",
        "patching",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_ignored_for_lower_priority_peer_when_already_off(
        k8s_mocked, assert_logs, settings, looptime,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 99,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    conflicts_found = aiotoggles.Toggle(False)
    stream_pressure = asyncio.Event()
    loop = asyncio.get_running_loop()
    loop.call_later(1.23, stream_pressure.set)

    assert conflicts_found.is_off()
    await process_peering_event(
        raw_event=event,
        conflicts_found=conflicts_found,
        stream_pressure=stream_pressure,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert conflicts_found.is_off()
    assert looptime == 0
    assert not k8s_mocked.patch.called
    assert_logs(prohibited=[
        "Possibly conflicting operators",
        "Pausing all operators, including self",
        "Pausing operations in favour of",
        "Resuming operations after the pause",
        "patching",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_toggled_on_for_same_priority_peer_when_initially_off(
        k8s_mocked, assert_logs, settings, looptime,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 100,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    conflicts_found = aiotoggles.Toggle(False)
    stream_pressure = asyncio.Event()
    loop = asyncio.get_running_loop()
    loop.call_later(1.23, stream_pressure.set)

    assert conflicts_found.is_off()
    await process_peering_event(
        raw_event=event,
        conflicts_found=conflicts_found,
        stream_pressure=stream_pressure,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert conflicts_found.is_on()
    assert looptime == 1.23
    assert not k8s_mocked.patch.called
    assert_logs([
        "Possibly conflicting operators",
        "Pausing all operators, including self",
    ], prohibited=[
        "Pausing operations in favour of",
        "Resuming operations after the pause",
        "patching",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_ignored_for_same_priority_peer_when_already_on(
        k8s_mocked, assert_logs, settings, looptime,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 100,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    conflicts_found = aiotoggles.Toggle(True)
    stream_pressure = asyncio.Event()
    loop = asyncio.get_running_loop()
    loop.call_later(1.23, stream_pressure.set)

    assert conflicts_found.is_on()
    await process_peering_event(
        raw_event=event,
        conflicts_found=conflicts_found,
        stream_pressure=stream_pressure,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert conflicts_found.is_on()
    assert looptime == 1.23
    assert not k8s_mocked.patch.called
    assert_logs([
        "Possibly conflicting operators",
    ], prohibited=[
        "Pausing all operators, including self",
        "Pausing operations in favour of",
        "Resuming operations after the pause",
        "patching",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
@pytest.mark.parametrize('priority', [100, 101])
async def test_resumes_immediately_on_expiration_of_blocking_peers(
        k8s_mocked, assert_logs, settings, priority, looptime,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': priority,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    conflicts_found = aiotoggles.Toggle(True)
    stream_pressure = asyncio.Event()

    assert conflicts_found.is_on()
    await process_peering_event(
        raw_event=event,
        conflicts_found=conflicts_found,
        stream_pressure=stream_pressure,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert conflicts_found.is_on()
    assert looptime == 9.876544
    assert k8s_mocked.patch.called
    assert_logs(prohibited=["patching"])



import asyncio
import json

import pytest

import kopf
from kopf._cogs.structs.ephemera import Memo
from kopf._core.engines.indexing import OperatorIndexers
from kopf._core.intents.causes import HANDLER_REASONS, Reason
from kopf._core.reactor.inventory import ResourceMemories
from kopf._core.reactor.processing import process_resource_event


@pytest.mark.parametrize('deletion_ts', [
    pytest.param({}, id='no-deletion-ts'),
    pytest.param({'deletionTimestamp': None}, id='empty-deletion-ts'),
    pytest.param({'deletionTimestamp': 'some'}, id='real-deletion-ts'),
])
@pytest.mark.parametrize('cause_type', HANDLER_REASONS)
async def test_1st_step_stores_progress_by_patching(
        registry, settings, handlers, extrahandlers,
        resource, cause_mock, cause_type, k8s_mocked, looptime, deletion_ts):
    name1 = f'{cause_type}_fn'
    name2 = f'{cause_type}_fn2'

    event_type = None if cause_type == Reason.RESUME else 'irrelevant'
    event_body = {
        'metadata': {'finalizers': [settings.persistence.finalizer]},
    }
    event_body['metadata'] |= deletion_ts
    cause_mock.reason = cause_type

    await process_resource_event(
        lifecycle=kopf.lifecycles.asap,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': event_body},
        event_queue=asyncio.Queue(),
    )

    assert handlers.create_mock.call_count == (1 if cause_type == Reason.CREATE else 0)
    assert handlers.update_mock.call_count == (1 if cause_type == Reason.UPDATE else 0)
    assert handlers.delete_mock.call_count == (1 if cause_type == Reason.DELETE else 0)
    assert handlers.resume_mock.call_count == (1 if cause_type == Reason.RESUME else 0)

    assert looptime == 0
    assert k8s_mocked.patch.called

    patch = k8s_mocked.patch.call_args_list[0].kwargs['payload']
    assert patch['metadata']['annotations']  # not empty at least
    progress1 = json.loads(patch['metadata']['annotations'][f"kopf.zalando.org/{name1}"])
    progress2 = json.loads(patch['metadata']['annotations'][f"kopf.zalando.org/{name2}"])

    assert progress1['retries'] == 1
    assert progress1['success'] is True

    assert progress2['retries'] == 0
    assert progress2['success'] is False

    assert progress1['started']
    assert progress2['started']

    # Premature removal of finalizers can prevent the 2nd step for deletion handlers.
    # So, the finalizers must never be removed on the 1st step.
    assert 'finalizers' not in patch['metadata']


@pytest.mark.parametrize('deletion_ts', [
    pytest.param({}, id='no-deletion-ts'),
    pytest.param({'deletionTimestamp': None}, id='empty-deletion-ts'),
    pytest.param({'deletionTimestamp': 'some'}, id='real-deletion-ts'),
])
@pytest.mark.parametrize('cause_type', HANDLER_REASONS)
async def test_2nd_step_finishes_the_handlers(caplog,
        registry, settings, handlers, extrahandlers,
        resource, cause_mock, cause_type, k8s_mocked, looptime, deletion_ts):
    name1 = f'{cause_type}_fn'
    name2 = f'{cause_type}_fn2'

    event_type = None if cause_type == Reason.RESUME else 'irrelevant'
    event_body = {
        'metadata': {
            'finalizers': [settings.persistence.finalizer],
            'resourceVersion': '1234567890',
            'annotations': {
                f'kopf.zalando.org/{name1}': json.dumps({
                    'started': '1979-01-01T00:00:00Z', 'success': True,
                }),
                f'kopf.zalando.org/{name2}': json.dumps({
                    'started': '1979-01-01T00:00:00Z'
                }),
            },
        },
        # 'status': {'kopf': {'progress': {
        #     name1: {'started': '1979-01-01T00:00:00Z', 'success': True},
        #     name2: {'started': '1979-01-01T00:00:00Z'},
        # }}}
    }
    event_body['metadata'] |= deletion_ts
    cause_mock.reason = cause_type

    await process_resource_event(
        lifecycle=kopf.lifecycles.one_by_one,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': event_body},
        event_queue=asyncio.Queue(),
    )

    assert extrahandlers.create_mock.call_count == (1 if cause_type == Reason.CREATE else 0)
    assert extrahandlers.update_mock.call_count == (1 if cause_type == Reason.UPDATE else 0)
    assert extrahandlers.delete_mock.call_count == (1 if cause_type == Reason.DELETE else 0)
    assert extrahandlers.resume_mock.call_count == (1 if cause_type == Reason.RESUME else 0)

    assert looptime == 0
    assert k8s_mocked.patch.called

    patch1 = k8s_mocked.patch.call_args_list[0].kwargs['payload']
    assert patch1['metadata']['annotations'][f'kopf.zalando.org/{name1}'] is None
    assert patch1['metadata']['annotations'][f'kopf.zalando.org/{name2}'] is None

    # Finalizers could be removed for resources being deleted on the 2nd step.
    # The logic can vary though: either by deletionTimestamp, or by reason==DELETE.
    if deletion_ts and deletion_ts['deletionTimestamp']:
        patch2 = k8s_mocked.patch.call_args_list[1].kwargs['payload']
        assert patch2 == [
            {'op': 'test', 'path': '/metadata/resourceVersion', 'value': '1234567890'},
            {'op': 'remove', 'path': '/metadata/finalizers'},
        ]



import asyncio
import dataclasses
import gc
import warnings

import pyngrok.conf
import pyngrok.ngrok
import pytest

from kopf._cogs.structs.references import Insights, Resource
from kopf._cogs.structs.reviews import CreateOptions, Request, RequestKind, RequestPayload, \
                                       RequestResource, UserInfo, WebhookFn
from kopf._core.engines.indexing import OperatorIndexers
from kopf._kits.webhooks import WebhookServer


# TODO: LATER: Fix this issue some day later.
@pytest.fixture()
def no_serverside_resource_warnings():
    """
    Hide an irrelevant ResourceWarning on the server side:

    It happens when a client disconnects from the webhook server,
    and the server closes the transport for that client. The garbage
    collector calls ``__del__()`` on the SSL proto object, despite
    it is not close to the moment.
    """
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore',
                                category=ResourceWarning,
                                module='asyncio.sslproto',
                                message='unclosed transport')
        yield

        # Provoke the garbage collection of SSL sockets to trigger the warnings.
        # Otherwise, in PyPy, these warnings leak to other tests due to delayed gc.
        gc.collect()


# TODO: LATER: Fix this issue after aiohttp 4.0.0 is used.
@pytest.fixture()
async def no_clientside_resource_warnings():
    """
    Hide an irrelevant ResourceWarning on the client side.

    https://docs.aiohttp.org/en/stable/client_advanced.html#graceful-shutdown
    """
    yield
    await asyncio.sleep(0.100)


@pytest.fixture()
async def no_sslproto_warnings(no_serverside_resource_warnings, no_clientside_resource_warnings):
    pass


# cert generation is somewhat slow (~1s)
@pytest.fixture(scope='module')
def certpkey():
    cert, pkey = WebhookServer.build_certificate(['localhost', '127.0.0.1'])
    return cert, pkey


@pytest.fixture()
def certfile(tmpdir, certpkey):
    path = tmpdir.join('cert.pem')
    path.write_binary(certpkey[0])
    return str(path)


@pytest.fixture()
def pkeyfile(tmpdir, certpkey):
    path = tmpdir.join('pkey.pem')
    path.write_binary(certpkey[1])
    return str(path)


@pytest.fixture()
def adm_request(resource, namespace):
    return Request(
        apiVersion='admission.k8s.io/v1',
        kind='AdmissionReview',
        request=RequestPayload(
            uid='uid1',
            kind=RequestKind(group=resource.group, version=resource.version, kind=resource.kind),
            resource=RequestResource(group=resource.group, version=resource.version, resource=resource.plural),
            subResource=None,
            requestKind=RequestKind(group=resource.group, version=resource.version, kind=resource.kind),
            requestResource=RequestResource(group=resource.group, version=resource.version, resource=resource.plural),
            requestSubResource=None,
            userInfo=UserInfo(username='user1', uid='useruid1', groups=['group1']),
            name='name1',
            namespace=namespace,
            operation='CREATE',
            options=CreateOptions(apiVersion='meta.k8s.io/v1', kind='CreateOptions'),
            object={'spec': {'field': 'value'}},
            oldObject=None,
            dryRun=False,
        ))


@dataclasses.dataclass(frozen=True)
class Responder:
    fn: WebhookFn
    fut: asyncio.Future  # asyncio.Future[Response]


@pytest.fixture()
async def responder() -> Responder:
    fut = asyncio.Future()
    async def fn(*_, **__):
        return await fut
    return Responder(fn=fn, fut=fut)


@pytest.fixture()
async def insights(settings, resource):
    val_resource = Resource('admissionregistration.k8s.io', 'v1', 'validatingwebhookconfigurations')
    mut_resource = Resource('admissionregistration.k8s.io', 'v1', 'mutatingwebhookconfigurations')
    insights = Insights()
    insights.watched_resources.add(resource)
    insights.webhook_resources.add(resource)
    await insights.backbone.fill(resources=[val_resource, mut_resource])
    insights.ready_resources.set()
    return insights


@pytest.fixture()
def indices():
    indexers = OperatorIndexers()
    return indexers.indices


@pytest.fixture(autouse=True)
def pyngrok_mock(mocker):
    mocker.patch.object(pyngrok.conf, 'get_default')
    mocker.patch.object(pyngrok.ngrok, 'set_auth_token')
    mocker.patch.object(pyngrok.ngrok, 'connect')
    mocker.patch.object(pyngrok.ngrok, 'disconnect')
    pyngrok.ngrok.connect.return_value.public_url = 'https://nowhere'
    return pyngrok



import aiohttp.web

from kopf._cogs.clients.auth import APIContext, authenticated
from kopf._cogs.structs.credentials import ConnectionInfo


@authenticated
async def fn(
        x: int,
        *,
        context: APIContext | None,
) -> tuple[APIContext, int]:
    return context, x + 100


async def test_session_is_injected(fake_vault, kmock, resource, namespace):
    kmock['get', resource, kmock.namespace(namespace), kmock.name('xyz')] << {}

    context, result = await fn(1)

    async with context.session:
        assert context is not None
        assert result == 101


async def test_session_is_passed_through(fake_vault, kmock, resource, namespace):
    kmock['get', resource, kmock.namespace(namespace), kmock.name('xyz')] << {}

    explicit_context = APIContext(ConnectionInfo(server='http://irrelevant/'))
    context, result = await fn(1, context=explicit_context)

    async with context.session:
        assert context is explicit_context
        assert result == 101



import asyncio
import collections.abc
import itertools
import json
import ssl
import urllib.parse
from collections.abc import AsyncIterator
from typing import Any

import aiohttp

from kopf._cogs.aiokits import aiotasks
from kopf._cogs.clients import auth, errors
from kopf._cogs.configs import configuration
from kopf._cogs.helpers import typedefs


@auth.authenticated
async def get_default_namespace(
        *,
        context: auth.APIContext | None = None,
) -> str | None:
    if context is None:
        raise RuntimeError("API instance is not injected by the decorator.")
    return context.default_namespace


@auth.authenticated
async def read_sslcert(
        *,
        context: auth.APIContext | None = None,
) -> tuple[str, bytes]:
    if context is None:
        raise RuntimeError("API instance is not injected by the decorator.")

    parsed = urllib.parse.urlparse(context.server)
    host = parsed.hostname or ''  # NB: it cannot be None/empty in our case.
    port = parsed.port or 443
    loop = asyncio.get_running_loop()
    cert = await loop.run_in_executor(None, ssl.get_server_certificate, (host, port))
    return host, cert.encode('ascii')


@auth.authenticated
async def request(
        method: str,
        url: str,  # relative to the server/api root.
        *,
        settings: configuration.OperatorSettings,
        payload: object | None = None,
        headers: dict[str, str] | None = None,
        timeout: aiohttp.ClientTimeout | None = None,
        context: auth.APIContext | None = None,  # injected by the decorator
        logger: typedefs.Logger,
) -> aiohttp.ClientResponse:
    if context is None:  # for type-checking!
        raise RuntimeError("API instance is not injected by the decorator.")

    if '://' not in url:
        url = context.server.rstrip('/') + '/' + url.lstrip('/')

    if timeout is None:
        timeout = aiohttp.ClientTimeout(
            total=settings.networking.request_timeout,
            sock_connect=settings.networking.connect_timeout,
        )

    backoffs = settings.networking.error_backoffs
    backoffs = backoffs if isinstance(backoffs, collections.abc.Iterable) else [backoffs]
    count = len(backoffs) + 1 if isinstance(backoffs, collections.abc.Sized) else None
    backoff: float | None
    for retry, backoff in enumerate(itertools.chain(backoffs, itertools.repeat(None)), start=1):
        idx = f"#{retry}/{count}" if count is not None else f"#{retry}"
        what = f"{method.upper()} {url}"
        try:
            if retry > 1:
                logger.debug(f"Request attempt {idx}: {what}")

            response = await context.session.request(
                method=method,
                url=url,
                json=payload,
                headers=headers,
                timeout=timeout,
            )
            await errors.check_response(response)  # but do not parse it!

        # aiohttp raises a generic error if the session/transport is closed, so we try to guess.
        # NB: "session closed" will reset the retry counter and do the full cycle with the new creds.
        except RuntimeError as e:
            if context.session.closed:
                # TODO: find a way to gracefully replace the active session in the existing context,
                #       so that all ongoing requests would switch to the new session & credentials.
                logger.error(f"Request attempt {idx} failed; TCP closed; will re-authenticate: {what}")
                raise errors.APISessionClosed("Session is closed.") from e
            raise

        # During k8s upgrades, API might throw 403 Forbidden. Use retries for this error as well.
        except (aiohttp.ClientConnectionError, errors.APIServerError, asyncio.TimeoutError,
                errors.APIForbiddenError, errors.APITooManyRequestsError) as e:

            # If we are asked to retry later, do so, and obey the requested backoff.
            if isinstance(e, errors.APITooManyRequestsError):
                if e.headers and e.headers.get("Retry-After"):
                    retry_after = int(float(e.headers["Retry-After"]))  # the new style
                elif e.details and e.details.get("retryAfterSeconds"):
                    retry_after = int(e.details["retryAfterSeconds"])  # the old style
                else:
                    retry_after = None

                if retry_after is not None and backoff is not None:
                    if settings.networking.enforce_retry_after or retry_after > backoff:
                        logger.debug(f"Overriding the backoff {backoff}s "
                                     f"with the retry-after {retry_after}s from the server: {what}")
                        backoff = retry_after

            if '[SSL: APPLICATION_DATA_AFTER_CLOSE_NOTIFY]' in str(e):  # for ClientOSError
                logger.error(f"Request attempt {idx} failed; SSL closed; will re-authenticate: {what}")
                raise errors.APISessionClosed("SSL data stream is closed.") from e
            elif backoff is None:  # the last or the only attempt.
                logger.error(f"Request attempt {idx} failed; escalating: {what} -> {e!r}")
                raise
            else:
                logger.error(f"Request attempt {idx} failed; will retry: {what} -> {e!r}")
                await asyncio.sleep(backoff)  # non-awakable! but still cancellable.
        else:
            if retry > 1:
                logger.debug(f"Request attempt {idx} succeeded: {what}")
            return response

    raise RuntimeError("Broken retryable routine.")  # impossible, but needed for type-checking.


async def get(
        url: str,  # relative to the server/api root.
        *,
        settings: configuration.OperatorSettings,
        payload: object | None = None,
        headers: dict[str, str] | None = None,
        timeout: aiohttp.ClientTimeout | None = None,
        logger: typedefs.Logger,
) -> Any:
    response = await request(
        method='get',
        url=url,
        payload=payload,
        headers=headers,
        timeout=timeout,
        settings=settings,
        logger=logger,
    )
    async with response:
        return await response.json()


async def post(
        url: str,  # relative to the server/api root.
        *,
        settings: configuration.OperatorSettings,
        payload: object | None = None,
        headers: dict[str, str] | None = None,
        timeout: aiohttp.ClientTimeout | None = None,
        logger: typedefs.Logger,
) -> Any:
    response = await request(
        method='post',
        url=url,
        payload=payload,
        headers=headers,
        timeout=timeout,
        settings=settings,
        logger=logger,
    )
    async with response:
        return await response.json()


async def patch(
        url: str,  # relative to the server/api root.
        *,
        settings: configuration.OperatorSettings,
        payload: object | None = None,
        headers: dict[str, str] | None = None,
        timeout: aiohttp.ClientTimeout | None = None,
        logger: typedefs.Logger,
) -> Any:
    response = await request(
        method='patch',
        url=url,
        payload=payload,
        headers=headers,
        timeout=timeout,
        settings=settings,
        logger=logger,
    )
    async with response:
        return await response.json()


async def delete(
        url: str,  # relative to the server/api root.
        *,
        settings: configuration.OperatorSettings,
        payload: object | None = None,
        headers: dict[str, str] | None = None,
        timeout: aiohttp.ClientTimeout | None = None,
        logger: typedefs.Logger,
) -> Any:
    response = await request(
        method='delete',
        url=url,
        payload=payload,
        headers=headers,
        timeout=timeout,
        settings=settings,
        logger=logger,
    )
    async with response:
        return await response.json()


async def stream(
        url: str,  # relative to the server/api root.
        *,
        settings: configuration.OperatorSettings,
        payload: object | None = None,
        headers: dict[str, str] | None = None,
        timeout: aiohttp.ClientTimeout | None = None,
        stopper: aiotasks.Future | None = None,
        logger: typedefs.Logger,
) -> AsyncIterator[Any]:
    # This dirty trickery is for cases when the server thinks too slowly before
    # sending the headers, but the stopper is already set during the initial wait.
    def request_cancel_callback(_: aiotasks.Future) -> None:
        task = asyncio.current_task()
        assert task is not None  # for type-checkers; this is `async def`, so always in a task.
        task.cancel()

    if stopper is not None and not stopper.done():
        stopper.add_done_callback(request_cancel_callback)
    try:
        response = await request(
            method='get',
            url=url,
            payload=payload,
            headers=headers,
            timeout=timeout,
            settings=settings,
            logger=logger,
        )
    except asyncio.CancelledError:
        if stopper is not None and stopper.done():
            return
        else:
            raise  # triggered not by the stopper, escalate
    finally:
        if stopper is not None:
            stopper.remove_done_callback(request_cancel_callback)

    # Do not proceed if managed to avoid the cancellation somehow, but got here.
    if stopper is not None and stopper.done():
        response.close()
        return

    # Once the headers were sent & received, the stopper works slightly differently:
    # it closes the response instead of cancelling the already performed request.
    def response_close_callback(_: aiotasks.Future) -> None:
        response.close()

    if stopper is not None:
        stopper.add_done_callback(response_close_callback)
    try:
        async with response:
            async for line in iter_jsonlines(response.content):
                yield json.loads(line.decode('utf-8'))
    except aiohttp.ClientConnectionError:
        if stopper is not None and stopper.done():
            pass
        else:
            raise  # triggered not by the stopper, escalate
    finally:
        if stopper is not None:
            stopper.remove_done_callback(response_close_callback)


async def iter_jsonlines(
        content: aiohttp.StreamReader,
        chunk_size: int = 1024 * 1024,
) -> AsyncIterator[bytes]:
    """
    Iterate line by line over the response's content.

    Usage:

    .. code-block:: python

        async for line in _iter_lines(response.content):
            pass

    This is an equivalent of:

    .. code-block:: python

        async for line in response.content:
            pass

    Except that the aiohttp's line iteration fails if the accumulated buffer
    length is above 2**17 bytes, i.e. 128 KB (`aiohttp.streams.DEFAULT_LIMIT`
    for the buffer's low-watermark, multiplied by 2 for the high-watermark).
    Kubernetes secrets and other fields can be much longer, up to MBs in length.

    The chunk size of 1MB is an empirical guess for keeping the memory footprint
    reasonably low on a huge number of small lines (limited to 1 MB in total),
    while ensuring the near-instant reads of the huge lines (can be a problem
    with a small chunk size due to too many iterations).

    .. seealso::
        https://github.com/zalando-incubator/kopf/issues/275
    """

    # Minimize the memory footprint by keeping at most 2 copies of a yielded line in memory
    # (in the buffer and as a yielded value), and at most 1 copy of other lines (in the buffer).
    buffer = b''
    async for data in content.iter_chunked(chunk_size):
        buffer += data
        del data

        start = 0
        index = buffer.find(b'\n', start)
        while index >= 0:
            line = buffer[start:index]
            if line:
                yield line
            del line
            start = index + 1
            index = buffer.find(b'\n', start)

        if start > 0:
            buffer = buffer[start:]

    if buffer:
        yield buffer



"""
Several webhooks servers & tunnels supported out of the box.
"""
import asyncio
import base64
import contextlib
import functools
import hashlib
import ipaddress
import json
import logging
import os
import pathlib
import socket
import ssl
import tempfile
import urllib.parse
from collections.abc import AsyncIterator, Collection, Iterable
from typing import TYPE_CHECKING, TypeAlias

import aiohttp.web

from kopf._cogs.clients import api, scanning
from kopf._cogs.configs import configuration
from kopf._cogs.structs import reviews
from kopf._core.engines import admission
from kopf._kits import webhacks

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    StrPath: TypeAlias = str | os.PathLike[str]
else:
    StrPath: TypeAlias = str | os.PathLike


class MissingDependencyError(ImportError):
    """ A server/tunnel is used which requires an optional dependency. """


class WebhookServer(webhacks.WebhookContextManager):
    """
    A local HTTP/HTTPS endpoint.

    Currently, the server is based on ``aiohttp``, but the implementation
    can change in the future without warning.

    This server is also used by specialised tunnels when they need
    a local endpoint to be tunneled.

    * ``addr``, ``port`` is where to listen for connections
      (defaults to ``localhost`` and ``9443``).
    * ``path`` is the root path for a webhook server
      (defaults to no root path).
    * ``host`` is an optional override of the hostname for webhook URLs;
      if not specified, the ``addr`` will be used.

    Kubernetes requires HTTPS, so HTTPS is the default mode of the server.
    This webhook server supports SSL both for the server certificates
    and for client certificates (e.g., for authentication) at the same time:

    * ``cadata``, ``cafile`` is the CA bundle to be passed as a "client config"
      to the webhook configuration objects, to be used by clients/apiservers
      when talking to the webhook server; it is not used in the server itself.
    * ``cadump`` is a path to save the resulting CA bundle to be used
      by clients, i.e. apiservers; it can be passed to ``curl --cacert ...``;
      if ``cafile`` is provided, it contains the same content.
    * ``certfile``, ``pkeyfile`` define the server's endpoint certificate;
      if not specified, a self-signed certificate and CA will be generated
      for both ``addr`` & ``host`` as SANs (but only ``host`` for CommonName).
    * ``password`` is either for decrypting the provided ``pkeyfile``,
      or for encrypting and decrypting the generated private key.
    * ``extra_sans`` are put into the self-signed certificate as SANs (DNS/IP)
      in addition to the host & addr (in case some other endpoints exist).
    * ``verify_mode``, ``verify_cafile``, ``verify_capath``, ``verify_cadata``
      will be loaded into the SSL context for verifying the client certificates
      when provided and if provided by the clients, i.e. apiservers or curl;
      (``ssl.SSLContext.verify_mode``, ``ssl.SSLContext.load_verify_locations``).
    * ``insecure`` flag disables HTTPS and runs an HTTP webhook server.
      This is used in ngrok for a local endpoint, but can be used for debugging
      or when the certificate-generating dependencies/extras are not installed.
    """
    DEFAULT_HOST: str | None = None

    addr: str | None  # None means "all interfaces"
    port: int | None  # None means random port
    host: str | None
    path: str | None

    cadata: bytes | None  # -> .webhooks.*.clientConfig.caBundle
    cafile: StrPath | None
    cadump: StrPath | None

    context: ssl.SSLContext | None
    insecure: bool
    certfile: StrPath | None
    pkeyfile: StrPath | None
    password: str | None

    extra_sans: Iterable[str]

    verify_mode: ssl.VerifyMode | None
    verify_cafile: StrPath | None
    verify_capath: StrPath | None
    verify_cadata: str | bytes | None

    file_check_interval: float

    def __init__(
            self,
            *,
            # Listening socket, root URL path, and the reported URL hostname:
            addr: str | None = None,
            port: int | None = None,
            path: str | None = None,
            host: str | None = None,
            # The CA bundle to be passed to "client configs":
            cadata: bytes | None = None,
            cafile: StrPath | None = None,
            cadump: StrPath | None = None,
            # A pre-configured SSL context (if any):
            context: ssl.SSLContext | None = None,
            # The server's own certificate, or lack of it (loaded into the context):
            insecure: bool = False,  # http is needed for ngrok
            certfile: StrPath | None = None,
            pkeyfile: StrPath | None = None,
            password: str | None = None,
            # Generated certificate's extra info.
            extra_sans: Iterable[str] = (),
            # Verification of client certificates (loaded into the context):
            verify_mode: ssl.VerifyMode | None = None,
            verify_cafile: StrPath | None = None,
            verify_capath: StrPath | None = None,
            verify_cadata: str | bytes | None = None,
            # File re-reading interval for cert renewals:
            file_check_interval: float = 60.,
    ) -> None:
        super().__init__()
        self.addr = addr
        self.port = port
        self.path = path
        self.host = host
        self.cadata = cadata
        self.cafile = cafile
        self.cadump = cadump
        self.context = context
        self.insecure = insecure
        self.certfile = certfile
        self.pkeyfile = pkeyfile
        self.password = password
        self.extra_sans = extra_sans
        self.verify_mode = verify_mode
        self.verify_cafile = verify_cafile
        self.verify_capath = verify_capath
        self.verify_cadata = verify_cadata
        self.file_check_interval = file_check_interval

    async def __call__(self, fn: reviews.WebhookFn) -> AsyncIterator[reviews.WebhookClientConfig]:

        # Redefine as a coroutine instead of a partial to avoid warnings from aiohttp.
        async def _serve_fn(request: aiohttp.web.Request) -> aiohttp.web.Response:
            return await self._serve(fn, request)

        while True:
            cadata, context = self._build_ssl()
            path = self.path.rstrip('/') if self.path else ''
            app = aiohttp.web.Application()
            app.add_routes([aiohttp.web.post(f"{path}/{{id:.*}}", _serve_fn)])
            runner = aiohttp.web.AppRunner(app, handle_signals=False)
            await runner.setup()
            try:
                # Note: reuse_port is mostly (but not only) for fast-running tests with SSL sockets;
                # multi-threaded sockets are not really used -- high load is not expected for webhooks.
                addr = self.addr or None  # None is aiohttp's "any interface"
                port = self.port or self._allocate_free_port()
                site = aiohttp.web.TCPSite(runner, addr, port, ssl_context=context, reuse_port=True, reuse_address=True)
                await site.start()

                # Log with the actual URL: normalised, with hostname/port set.
                schema = 'http' if context is None else 'https'
                url = self._build_url(schema, addr or '*', port, self.path or '')
                logger.debug(f"Listening for webhooks at {url}")
                host = self.host or self.DEFAULT_HOST or self._get_accessible_addr(self.addr)
                url = self._build_url(schema, host, port, self.path or '')
                logger.debug(f"Accessing the webhooks at {url}")

                client_config = reviews.WebhookClientConfig(url=url)
                if cadata is not None:
                    client_config['caBundle'] = base64.b64encode(cadata).decode('ascii')

                yield client_config
                await self.__sleep_forever_or_until_ssl_files_change()
                logger.debug(f"Restarting with the renewed SSL certificate files.")
            finally:
                # On any reason of exit, stop serving the endpoint.
                await runner.cleanup()

    @staticmethod
    async def _serve(
            fn: reviews.WebhookFn,
            request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        """
        Serve a single admission request: an aiohttp-specific implementation.

        Mind 2 different ways the errors are reported:

        * Directly by the webhook's response, i.e. to the apiservers.
          This means that the webhook request was done improperly;
          the original API request might be good, but we could not confirm that.
        * In ``.response.status``, as apiservers send it to the requesting user.
          This means that the original API operation was done improperly,
          while the webhooks are functional.
        """
        # The extra information that is passed down to handlers for authentication/authorization.
        # Note: this is an identity of an apiserver, not of the user that sends an API request.
        headers = dict(request.headers)
        sslpeer = request.transport.get_extra_info('peercert') if request.transport else None
        webhook = request.match_info.get('id')
        try:
            text = await request.text()
            data = json.loads(text)
            response = await fn(data, webhook=webhook, sslpeer=sslpeer, headers=headers)
            return aiohttp.web.json_response(response)
        except admission.AmbiguousResourceError as e:
            raise aiohttp.web.HTTPConflict(reason=str(e) or None)
        except admission.UnknownResourceError as e:
            raise aiohttp.web.HTTPNotFound(reason=str(e) or None)
        except admission.WebhookError as e:
            raise aiohttp.web.HTTPBadRequest(reason=str(e) or None)
        except json.JSONDecodeError as e:
            raise aiohttp.web.HTTPBadRequest(reason=str(e) or None)

    @staticmethod
    def _allocate_free_port() -> int:
        with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            s.bind(('', 0))  # '' is a special IPv4 form for "any interface"
            return int(s.getsockname()[1])

    @staticmethod
    def _get_accessible_addr(addr: str | None) -> str:
        """
        Convert a "catch-all" listening address to the accessible hostname.

        "Catch-all" interfaces like ``0.0.0.0`` or ``::/0`` can be used
        for listening to utilise all interfaces, but cannot be accessed.
        Some other real ("specified") address must be used for that.

        If the address is not IPv4/IPv6 address or is a regular "specified"
        address, it is used as is. Only the special addressed are overridden.
        """
        if addr is None:
            return 'localhost'  # and let the system resolved it to IPv4/IPv6
        try:
            ipv4 = ipaddress.IPv4Address(addr)
        except ipaddress.AddressValueError:
            pass
        else:
            return '127.0.0.1' if ipv4.is_unspecified else addr
        try:
            ipv6 = ipaddress.IPv6Address(addr)
        except ipaddress.AddressValueError:
            pass
        else:
            return '::1' if ipv6.is_unspecified else addr
        return addr

    @staticmethod
    def _build_url(schema: str, host: str, port: int, path: str) -> str:
        try:
            ipv6 = ipaddress.IPv6Address(host)
        except ipaddress.AddressValueError:
            pass
        else:
            host = f'[{ipv6}]'
        is_default_port = ((schema == 'http' and port == 80) or
                           (schema == 'https' and port == 443))
        netloc = host if is_default_port else f'{host}:{port}'
        return urllib.parse.urlunsplit([schema, netloc, path, '', ''])

    def _build_ssl(self) -> tuple[bytes | None, ssl.SSLContext | None]:
        """
        A macro to construct an SSL context, possibly generating SSL certs.

        Returns a CA bundle to be passed to the "client configs",
        and a properly initialised SSL context to be used by the server.
        Or ``None`` for both if an HTTP server is needed.
        """
        cadata = self.cadata
        context = self.context
        if self.insecure and self.context is not None:
            raise ValueError("Insecure mode cannot have an SSL context specified.")

        # Read the provided CA bundle for webhooks' "client config"; not used by the server itself.
        if cadata is None and self.cafile is not None:
            cadata = pathlib.Path(self.cafile).read_bytes()

        # Kubernetes does not work with HTTP, so we do not bother and always run HTTPS too.
        # Except when explicitly said to be insecure, e.g. by ngrok (free plan only supports HTTP).
        if context is None and not self.insecure:
            context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)

        if context is not None:

            # Load a CA for verifying the client certificates (if provided) by this server.
            if self.verify_mode is not None:
                context.verify_mode = self.verify_mode
            if self.verify_cafile or self.verify_capath or self.verify_cadata:
                logger.debug("Loading a CA for client certificate verification.")
                context.load_verify_locations(
                    self.verify_cafile,
                    self.verify_capath,
                    self.verify_cadata,
                )
                if context.verify_mode == ssl.CERT_NONE:
                    context.verify_mode = ssl.CERT_OPTIONAL

            # Load the specified server's certificate, or generate a self-signed one if possible.
            # If cafile/cadata are not defined, use the server's certificate as a CA for clients.
            if self.certfile is not None and self.pkeyfile is not None:
                logger.debug("Using a provided certificate for HTTPS.")
                context.load_cert_chain(
                    self.certfile,
                    self.pkeyfile,
                    self.password,
                )
                if cadata is None and self.certfile is not None:
                    cadata = pathlib.Path(self.certfile).read_bytes()
            else:
                logger.debug("Generating a self-signed certificate for HTTPS.")
                host = self.host or self.DEFAULT_HOST
                addr = self._get_accessible_addr(self.addr)
                hostnames = [host or addr, addr] + list(self.extra_sans)
                certdata, pkeydata = self.build_certificate(hostnames, self.password)
                with tempfile.NamedTemporaryFile() as certf, tempfile.NamedTemporaryFile() as pkeyf:
                    certf.write(certdata)
                    pkeyf.write(pkeydata)
                    certf.flush()
                    pkeyf.flush()
                    context.load_cert_chain(certf.name, pkeyf.name, self.password)

                # For a self-signed certificate, the CA bundle is the certificate itself,
                # regardless of what cafile/cadata are provided from outside.
                cadata = certdata

        # Dump the provided or self-signed CA (but not the key!), e.g. for `curl --cacert ...`
        if self.cadump is not None and cadata is not None:
            pathlib.Path(self.cadump).write_bytes(cadata)

        return cadata, context

    async def __sleep_forever_or_until_ssl_files_change(self) -> None:
        """
        Wake up every time the SSL cert files change. Sleep forever otherwise.

        The mounted secrets do not renew their metadata or send the filesystem
        events, so only a full re-read is the guaranteed way of file monitoring.
        """
        paths = [self.certfile, self.pkeyfile, self.cafile]
        paths = [path for path in paths if path is not None]
        initial_digest: str | None = None
        while True:
            # For security, do not keep the secrets in memory, hash & garbage collect them asap.
            # Also, minimize the memory footprint by keeping only a tiny hash of all files combined.
            hasher = hashlib.new('sha256')  # the algorith is irrelevant
            for path in paths:
                with open(str(path), 'rb') as f:
                    hasher.update(f.read())
            digest = hasher.hexdigest()

            if initial_digest is None:
                initial_digest = digest

            if digest == initial_digest:
                await asyncio.sleep(self.file_check_interval)
            else:
                return

    @staticmethod
    def build_certificate(
            hostnames: Collection[str],
            password: str | None = None,
    ) -> tuple[bytes, bytes]:
        """
        Build a self-signed certificate with SANs (subject alternative names).

        Returns a tuple of the certificate and its private key (PEM-formatted).

        The certificate is "minimally sufficient", without much of the extra
        information on the subject besides its common and alternative names.
        However, IP addresses are properly recognised and normalised for better
        compatibility with strict SSL clients (like apiservers of Kubernetes).
        The first non-IP hostname becomes the certificate's common name --
        by convention, non-configurable. If no hostnames are found, the first
        IP address is used as a fallback. Magic IPs like 0.0.0.0 are excluded.

        ``certbuilder`` is used as an implementation because it is lightweight:
        2.9 MB vs. 8.7 MB for cryptography. Still, it is too heavy to include
        as a normal runtime dependency (for 8.8 MB of Kopf itself), so it is
        only available as the ``kopf[dev]`` extra for development-mode dependencies.
        This can change in the future if self-signed certificates become used
        at runtime (e.g. in production/staging environments or other real clusters).
        """
        try:
            import certbuilder
            import oscrypto.asymmetric
        except ImportError:
            raise MissingDependencyError(
                "Using self-signed certificates requires an extra dependency: "
                "run `pip install certbuilder` or `pip install kopf[dev]`. "
                "Or pass `insecure=True` to a webhook server to use only HTTP. "
                "Or generate your own certificates and pass as certfile=/pkeyfile=. "
                "More: https://docs.kopf.dev/en/stable/admission/")

        # Detect which ones of the hostnames are probably IPv4/IPv6 addresses.
        # A side-effect: bring them all to their canonical forms.
        parsed_ips: dict[str, ipaddress.IPv4Address | ipaddress.IPv6Address] = {}
        for hostname in hostnames:
            try:
                parsed_ips[hostname] = ipaddress.IPv4Address(hostname)
            except ipaddress.AddressValueError:
                pass  # non-parsable IPs are considered to be regular hostnames
            try:
                parsed_ips[hostname] = ipaddress.IPv6Address(hostname)
            except ipaddress.AddressValueError:
                pass  # non-parsable IPs are considered to be regular hostnames

        # Later, only the normalised IPs are used as SANs, not the raw IPs.
        # Remove bindable but non-accessible addresses (like 0.0.0.0) form the SANs.
        true_hostnames = [hostname for hostname in hostnames if hostname not in parsed_ips]
        accessible_ips = [str(ip) for ip in parsed_ips.values() if not ip.is_unspecified]

        # Build a certificate as the framework believe is good enough for itself.
        subject = {'common_name': true_hostnames[0] if true_hostnames else accessible_ips[0]}
        public_key, private_key = oscrypto.asymmetric.generate_pair('rsa', bit_size=2048)
        builder = certbuilder.CertificateBuilder(subject, public_key)
        builder.ca = True
        builder.key_usage = {'digital_signature', 'key_encipherment', 'key_cert_sign', 'crl_sign'}
        builder.extended_key_usage = {'server_auth', 'client_auth'}
        builder.self_signed = True
        builder.subject_alt_ips = list(set(accessible_ips))  # deduplicate
        builder.subject_alt_domains = list(set(true_hostnames) | set(accessible_ips))  # deduplicate
        certificate = builder.build(private_key)
        cert_pem: bytes = certbuilder.pem_armor_certificate(certificate)
        pkey_pem: bytes = oscrypto.asymmetric.dump_private_key(private_key, password, target_ms=10)
        return cert_pem, pkey_pem


class WebhookK3dServer(WebhookServer):
    """
    A tunnel from inside of K3d/K3s to its host where the operator is running.

    With this tunnel, a developer can develop the webhooks when fully offline,
    since all the traffic is local and never leaves the host machine.

    The forwarding is maintained by K3d itself. This tunnel only replaces
    the endpoints for the Kubernetes webhook and injects an SSL certificate
    with proper CN/SANs --- to match Kubernetes's SSL validity expectations.
    """
    DEFAULT_HOST = 'host.k3d.internal'


class WebhookMinikubeServer(WebhookServer):
    """
    A tunnel from inside of Minikube to its host where the operator is running.

    With this tunnel, a developer can develop the webhooks when fully offline,
    since all the traffic is local and never leaves the host machine.

    The forwarding is maintained by Minikube itself. This tunnel only replaces
    the endpoints for the Kubernetes webhook and injects an SSL certificate
    with proper CN/SANs --- to match Kubernetes's SSL validity expectations.
    """
    DEFAULT_HOST = 'host.minikube.internal'


class WebhookDockerDesktopServer(WebhookServer):
    """A tunnel from inside of Docker Desktop to its host where the operator is
    running.

    With this tunnel, a developer can develop the webhooks when fully
    offline, since all the traffic is local and never leaves the host
    machine.

    The forwarding is maintained by Docker Desktop itself. This tunnel
    only replaces the endpoints for the Kubernetes webhook and injects
    an SSL certificate with proper CN/SANs --- to match Kubernetes's SSL
    validity expectations.
    """
    DEFAULT_HOST = "host.docker.internal"


class WebhookNgrokTunnel(webhacks.WebhookContextManager):
    """
    Tunnel admission webhook requests via an external tunnel: ngrok_.

    .. _ngrok: https://ngrok.com/

    ``addr``, ``port``, and ``path`` have the same meaning as in
    :class:`kopf.WebhookServer`: where to listen for connections locally.
    Ngrok then tunnels this endpoint to a remote public URL.

    Mind that the ngrok webhook tunnel runs the local webhook server
    in an insecure (HTTP) mode. For secure (HTTPS) mode, a paid subscription
    and properly issued certificates are needed. This goes beyond Kopf's scope.
    If needed, implement your own ngrok tunnel.

    Besides, ngrok tunnel does not report any CA to the webhook client configs.
    It is expected that the default trust chain is sufficient for ngrok's certs.

    ``token`` can be used for paid subscriptions, which lifts some limitations.
    Otherwise, the free plan has a limit of 40 requests per minute
    (this should be enough for local development).

    ``binary``, if set, will use the specified ``ngrok`` binary path;
    otherwise, ``pyngrok`` downloads the binary at runtime (not recommended).

    .. warning::

        The public URL is not properly protected and a malicious user
        can send requests to a locally running operator. If the handlers
        only process the data and make no side effects, this should be fine.

        Despite ngrok providing basic auth ("username:password"),
        Kubernetes does not permit this information in the URLs.

        Ngrok partially "protects" the URLS by assigning them random hostnames.
        Additionally, you can add random paths. However, this is not "security",
        only a bit of safety for a short time (enough for development runs).
    """
    addr: str | None  # None means "any interface"
    port: int | None  # None means a random port
    path: str | None
    token: str | None
    region: str | None
    binary: StrPath | None

    def __init__(
            self,
            *,
            addr: str | None = None,
            port: int | None = None,
            path: str | None = None,
            token: str | None = None,
            region: str | None = None,
            binary: StrPath | None = None,
    ) -> None:
        super().__init__()
        self.addr = addr
        self.port = port
        self.path = path
        self.token = token
        self.region = region
        self.binary = binary

    async def __call__(self, fn: reviews.WebhookFn) -> AsyncIterator[reviews.WebhookClientConfig]:
        try:
            from pyngrok import conf, ngrok
        except ImportError:
            raise MissingDependencyError(
                "Using ngrok webhook tunnel requires an extra dependency: "
                "run `pip install pyngrok` or `pip install kopf[dev]`. "
                "More: https://docs.kopf.dev/en/stable/admission/")

        if self.binary is not None:
            conf.get_default().ngrok_path = str(self.binary)
        if self.region is not None:
            conf.get_default().region = self.region
        if self.token is not None:
            ngrok.set_auth_token(self.token)

        # Ngrok only supports HTTP with a free plan; HTTPS requires a paid subscription.
        tunnel: ngrok.NgrokTunnel | None = None
        loop = asyncio.get_running_loop()
        async with WebhookServer(addr=self.addr, port=self.port,
                                 path=self.path, insecure=True) as server:
            # TODO: inverse try & async for?
            try:
                async for client_config in server(fn):

                    # Re-create the tunnel for each new local endpoint (if it did change at all).
                    if tunnel is not None:
                        await loop.run_in_executor(None, ngrok.disconnect, tunnel.public_url)
                    parsed = urllib.parse.urlparse(client_config['url'])
                    tunnel = await loop.run_in_executor(
                        None, functools.partial(ngrok.connect, f'{parsed.port}', bind_tls=True))

                    # Adjust for local webhook server specifics (no port, but with the same path).
                    # Report no CA bundle -- ngrok's certs (Let's Encrypt) are in a default trust chain.
                    assert tunnel is not None  # for mypy
                    url = f"{tunnel.public_url}{self.path or ''}"
                    logger.debug(f"Accessing the webhooks at {url}")
                    yield reviews.WebhookClientConfig(url=url)  # e.g. 'https://e5fc05f6494b.ngrok.io/xyz'
            finally:
                if tunnel is not None:
                    await loop.run_in_executor(None, ngrok.disconnect, tunnel.public_url)


class ClusterDetector:
    """
    A mixin for auto-server/auto-tunnel to detect the cluster type.

    The implementation of the server detection requires the least possible
    permissions or no permissions at all. In most cases, it will identify
    the server type by its SSL certificate meta-information (subject/issuer).
    SSL information is the most universal way for all typical local clusters.

    If SSL parsing fails, it will try to fetch the information from the cluster.
    However, it rarely contains any useful information about the cluster's
    surroundings and environment, but only about the cluster itself
    (though it helps with K3s).

    Note: the SSL certificate of the Kubernetes API is checked, not of webhooks.
    """
    @staticmethod
    async def guess_host() -> str | None:
        try:
            import certvalidator
        except ImportError:
            raise MissingDependencyError(
                "Auto-guessing cluster types requires an extra dependency: "
                "run `pip install certvalidator` or `pip install kopf[dev]`. "
                "More: https://docs.kopf.dev/en/stable/admission/")

        hostname, cert = await api.read_sslcert()
        valcontext = certvalidator.ValidationContext(extra_trust_roots=[cert])
        validator = certvalidator.CertificateValidator(cert, validation_context=valcontext)
        certpath = validator.validate_tls(hostname)
        issuer_cn = certpath.first.issuer.native.get('common_name', '')
        subject_cn = certpath.first.subject.native.get('common_name', '')
        subject_org = certpath.first.subject.native.get('organization_name', '')
        subject_alt_names = [name for name in certpath.first.subject_alt_name_value.native]

        if subject_cn == 'k3s' or subject_org == 'k3s' or issuer_cn.startswith('k3s-'):
            return WebhookK3dServer.DEFAULT_HOST
        elif subject_cn == 'minikube' or issuer_cn == 'minikubeCA':
            return WebhookMinikubeServer.DEFAULT_HOST
        elif 'docker-for-desktop' in subject_alt_names:
            return WebhookDockerDesktopServer.DEFAULT_HOST
        else:
            # The default timeouts & backoffs are used to retrieve the cluster
            # version, not those from the operator. It is too difficult to get
            # the settings here in webhooks. The "proper" way is to retrieve
            # the version in observation routines and pass it via contextvars,
            # but this is too overcomplicated for a dev-mode helping utility.
            settings = configuration.OperatorSettings()
            versioninfo = await scanning.read_version(settings=settings, logger=logger)
            if '+k3s' in versioninfo.get('gitVersion', ''):
                return WebhookK3dServer.DEFAULT_HOST
        return None


class WebhookAutoServer(ClusterDetector, WebhookServer):
    """
    A locally listening webserver which attempts to guess its proper hostname.

    The choice is happening between supported webhook servers only
    (K3d/K3d and Minikube at the moment). In all other cases,
    a regular local server is started without hostname overrides.

    If automatic tunneling is possible, consider :class:`WebhookAutoTunnel`.
    """
    async def __call__(self, fn: reviews.WebhookFn) -> AsyncIterator[reviews.WebhookClientConfig]:
        host = self.DEFAULT_HOST = await self.guess_host()
        if host is None:
            logger.debug(f"Cluster detection failed, running a simple local server.")
        else:
            logger.debug(f"Cluster detection found the hostname: {host}")
        async for client_config in super().__call__(fn):
            yield client_config


class WebhookAutoTunnel(ClusterDetector, webhacks.WebhookContextManager):
    """
    The same as :class:`WebhookAutoServer`, but with possible tunneling.

    Generally, tunneling gives more possibilities to run in any environment,
    but it must not happen without a permission from the developers,
    and is not possible if running in a completely isolated/local/CI/CD cluster.
    Therefore, developers should activate automatic setup explicitly.

    If automatic tunneling is prohibited or impossible,
    use :class:`WebhookAutoServer`.

    .. note::

        Automatic server/tunnel detection is highly limited in configuration
        and provides only the most common options of all servers & tunnels:
        specifically, listening ``addr:port/path``.
        All other options are specific to their servers/tunnels,
        and the auto-guessing logic cannot use/accept/pass them.
    """
    addr: str | None  # None means "any interface"
    port: int | None  # None means a random port
    path: str | None

    def __init__(
            self,
            *,
            addr: str | None = None,
            port: int | None = None,
            path: str | None = None,
    ) -> None:
        super().__init__()
        self.addr = addr
        self.port = port
        self.path = path

    async def __call__(self, fn: reviews.WebhookFn) -> AsyncIterator[reviews.WebhookClientConfig]:
        server: WebhookNgrokTunnel | WebhookServer
        host = await self.guess_host()
        if host is None:
            logger.debug(f"Cluster detection failed, using an ngrok tunnel.")
            server = WebhookNgrokTunnel(addr=self.addr, port=self.port, path=self.path)
        else:
            logger.debug(f"Cluster detection found the hostname: {host}")
            server = WebhookServer(addr=self.addr, port=self.port, path=self.path, host=host)
        async with server:
            async for client_config in server(fn):
                yield client_config



import asyncio

import kmock
import pytest

import kopf
from kopf._cogs.structs.bodies import RawBody, RawEvent
from kopf._cogs.structs.references import NAMESPACES, Insights, Resource
from kopf._core.reactor.observation import process_discovered_resource_event

# Implementation awareness: the events only trigger the re-scan, so the fields can be reduced
# to only the group name which is being rescanned. Other fields are ignored in the events.
# The actual data is taken from the API. It is tested elsewhere, so we rely on its correctness here.
GROUP1_EVENT = RawBody(spec={'group': 'group1'})

# Since we are affecting the K8s API discovery, implicit URL handling is a problem. Disable it.
pytestmark = [
    pytest.mark.kmock(cls=kmock.RawHandler),
    pytest.mark.usefixtures('fake_vault'),
]


@pytest.fixture()
def core_mock(kmock):
    return kmock['get /api'] << {'versions': ['v1']}


@pytest.fixture()
def apis_mock(kmock):
    return kmock['get /apis'] << {'groups': [
        {
            'name': 'group1',
            'preferredVersion': {'version': 'version1'},
            'versions': [{'version': 'version1'}],
        },
    ]}


@pytest.fixture()
def corev1_mock(kmock, core_mock):
    return kmock['get /api/v1'] << {'resources': [
        {
            'kind': 'Namespace',
            'name': 'namespaces',
            'singularName': 'namespace',
            'namespaced': False,
            'categories': [],
            'shortNames': [],
            'verbs': ['list', 'watch', 'patch'],
        },
    ]}


@pytest.fixture()
def group1_mock(kmock, apis_mock):
    return kmock['get /apis/group1/version1'] << {'resources': [
        {
            'kind': 'kind1',
            'name': 'plural1',
            'singularName': 'singular1',
            'namespaced': True,
            'categories': ['category1a', 'category1b'],
            'shortNames': ['shortname1a', 'shortname1b'],
            'verbs': ['list', 'watch', 'patch'],
        },
    ]}


@pytest.fixture()
def group1_empty_mock(kmock, apis_mock):
    return kmock['get /apis/group1/version1'] << {'resources': []}


@pytest.fixture()
def group1_404mock(kmock, apis_mock):
    return kmock['get /apis/group1/version1'] << 404


@pytest.fixture()
async def insights():
    return Insights()


@pytest.fixture(params=[
    (kopf.on.event, 'watched_resources'),
    (kopf.daemon, 'watched_resources'),
    (kopf.timer, 'watched_resources'),
    (kopf.index, 'watched_resources'),
    (kopf.index, 'indexed_resources'),
    (kopf.on.resume, 'watched_resources'),
    (kopf.on.create, 'watched_resources'),
    (kopf.on.update, 'watched_resources'),
    (kopf.on.delete, 'watched_resources'),
    (kopf.on.validate, 'webhook_resources'),
    (kopf.on.mutate, 'webhook_resources'),
])
def insights_resources(request, registry, insights):
    decorator, insights_field = request.param

    @decorator('group1', 'version1', 'plural1')
    def fn(**_): pass

    return getattr(insights, insights_field)


@pytest.mark.parametrize('decorator', [kopf.on.validate, kopf.on.mutate])
@pytest.mark.parametrize('etype', ['ADDED', 'MODIFIED'])
async def test_nonwatchable_resources_are_ignored(
        settings, registry, apis_mock, group1_mock, looptime, etype, decorator, insights):

    @decorator('group1', 'version1', 'plural1')
    def fn(**_): pass

    e1 = RawEvent(type=etype, object=RawBody(spec={'group': 'group1'}))

    async def delayed_injection(delay: float):
        await asyncio.sleep(delay)
        await process_discovered_resource_event(
            insights=insights, raw_event=e1, registry=registry, settings=settings)

    task = asyncio.create_task(delayed_injection(1.23))
    async with insights.revised:
        await insights.revised.wait()
    await task
    assert looptime == 1.23
    assert not insights.watched_resources
    assert len(apis_mock) > 0
    assert len(group1_mock) > 0


async def test_initial_listing_is_ignored(
        settings, registry, apis_mock, group1_mock, looptime, insights):

    e1 = RawEvent(type=None, object=RawBody(spec={'group': 'group1'}))

    async def delayed_injection(delay: float):
        await asyncio.sleep(delay)
        await process_discovered_resource_event(
            insights=insights, raw_event=e1, registry=registry, settings=settings)

    task = asyncio.create_task(delayed_injection(0))
    with pytest.raises(asyncio.TimeoutError):
        async with insights.revised:
            await asyncio.wait_for(insights.revised.wait(), timeout=1.23)
    await task

    assert looptime == 1.23
    assert not insights.indexed_resources
    assert not insights.watched_resources
    assert not insights.webhook_resources
    assert not len(apis_mock) > 0
    assert not len(group1_mock) > 0


@pytest.mark.parametrize('etype', ['ADDED', 'MODIFIED'])
async def test_followups_for_addition(
        settings, registry, apis_mock, group1_mock, looptime, etype, insights, insights_resources):

    e1 = RawEvent(type=etype, object=RawBody(spec={'group': 'group1'}))
    r1 = Resource(group='group1', version='version1', plural='plural1')

    async def delayed_injection(delay: float):
        await asyncio.sleep(delay)
        await process_discovered_resource_event(
            insights=insights, raw_event=e1, registry=registry, settings=settings)

    task = asyncio.create_task(delayed_injection(1.23))
    async with insights.revised:
        await insights.revised.wait()
    await task

    assert looptime == 1.23
    assert insights_resources == {r1}
    assert len(apis_mock) > 0
    assert len(group1_mock) > 0


@pytest.mark.parametrize('etype', ['ADDED', 'MODIFIED', 'DELETED'])
async def test_followups_for_deletion_of_resource(
        settings, registry, apis_mock, group1_empty_mock, looptime, etype,
        insights, insights_resources):

    e1 = RawEvent(type=etype, object=RawBody(spec={'group': 'group1'}))
    r1 = Resource(group='group1', version='version1', plural='plural1')
    insights_resources.add(r1)

    async def delayed_injection(delay: float):
        await asyncio.sleep(delay)
        await process_discovered_resource_event(
            insights=insights, raw_event=e1, registry=registry, settings=settings)

    task = asyncio.create_task(delayed_injection(1.23))
    async with insights.revised:
        await insights.revised.wait()
    await task

    assert looptime == 1.23
    assert not insights_resources
    assert len(apis_mock) > 0
    assert len(group1_empty_mock) > 0


@pytest.mark.parametrize('etype', ['ADDED', 'MODIFIED', 'DELETED'])
async def test_followups_for_deletion_of_group(
        settings, registry, apis_mock, group1_404mock, looptime, etype, insights, insights_resources):

    e1 = RawEvent(type=etype, object=RawBody(spec={'group': 'group1'}))
    r1 = Resource(group='group1', version='version1', plural='plural1')
    insights_resources.add(r1)

    async def delayed_injection(delay: float):
        await asyncio.sleep(delay)
        await process_discovered_resource_event(
            insights=insights, raw_event=e1, registry=registry, settings=settings)

    task = asyncio.create_task(delayed_injection(1.23))
    async with insights.revised:
        await insights.revised.wait()
    await task

    assert looptime == 1.23
    assert not insights_resources
    assert len(apis_mock) > 0
    assert len(group1_404mock) > 0


@pytest.mark.parametrize('etype', ['DELETED'])
async def test_backbone_is_filled(
        settings, registry, core_mock, corev1_mock, looptime, etype, insights):

    e1 = RawEvent(type=etype, object=RawBody(spec={'group': ''}))

    async def delayed_injection(delay: float):
        await asyncio.sleep(delay)
        await process_discovered_resource_event(
            insights=insights, raw_event=e1, registry=registry, settings=settings)

    task = asyncio.create_task(delayed_injection(1.23))
    await insights.backbone.wait_for(NAMESPACES)
    await task

    assert looptime == 1.23
    assert NAMESPACES in insights.backbone
    assert len(core_mock) > 0
    assert len(corev1_mock) > 0



from kopf._core.intents.registries import ActivityRegistry, OperatorRegistry, ResourceRegistry


def test_activity_registry(activity_registry_cls):
    registry = activity_registry_cls()
    assert isinstance(registry, ActivityRegistry)


def test_resource_registry(resource_registry_cls):
    registry = resource_registry_cls()
    assert isinstance(registry, ResourceRegistry)


def test_operator_registry(operator_registry_cls):
    registry = operator_registry_cls()
    assert isinstance(registry, OperatorRegistry)
    assert not isinstance(registry, ResourceRegistry)



from itertools import chain, repeat
from unittest import mock

import pytest

from kopf._core.engines.peering import keepalive


class StopInfiniteCycleException(Exception):
    pass


async def test_background_task_runs(mocker, settings, assert_logs, namespaced_peering_resource):
    touch_mock = mocker.patch('kopf._core.engines.peering.touch')

    sleep_mock = mocker.patch('asyncio.sleep')
    # restore the default behavior after exhausting test values.
    # pytest-aiohttp calls asyncio.sleep during teardown, before the mock is removed.
    sleep_mock.side_effect = chain([None, None, StopInfiniteCycleException], repeat(mock.DEFAULT))

    randint_mock = mocker.patch('random.randint')
    randint_mock.side_effect = [7, 5, 9]

    settings.peering.lifetime = 33
    with pytest.raises(StopInfiniteCycleException):
        await keepalive(settings=settings, identity='id',
                        resource=namespaced_peering_resource, namespace='namespace')

    assert randint_mock.call_count == 3  # only to be sure that we test the right thing
    assert sleep_mock.call_count == 3
    assert sleep_mock.call_args_list[0].args[0] == 33 - 7
    assert sleep_mock.call_args_list[1].args[0] == 33 - 5
    assert sleep_mock.call_args_list[2].args[0] == 33 - 9

    assert touch_mock.call_count == 4  # 3 updates + 1 clean-up
    assert_logs(prohibited=["patching"])



"""
Supporting tasks for startup/cleanup and to keep the operator functional.

Consumes a credentials vault, and monitors that it has enough credentials.
When the credentials are invalidated (i.e. excluded), run the re-authentication
activity and populate them with new credentials (fully or partially).

The process is intentionally split into multiple packages:

* Authenticating background task (this module) is a part of the reactor,
  as it will not be able to run without up-to-date credentials,
  and since it initiates the reactor's activities and invokes the handlers.
* Vault are the data structures used mostly in the API clients wrappers
  (which are the low-level modules, so they cannot import the credentials
  from the high-level modules such as the reactor/engines).
* Specific authentication methods, such as the authentication piggybacking,
  belong to neither the reactor, nor the engines, nor the client wrappers.
"""
import logging
from typing import NoReturn

from kopf._cogs.aiokits import aiotime
from kopf._cogs.configs import configuration
from kopf._cogs.structs import credentials, ephemera, ids
from kopf._core.actions import execution, lifecycles, progression
from kopf._core.intents import causes, registries

logger = logging.getLogger(__name__)


class ActivityError(Exception):
    """ An error in the activity, caused by mandatory handlers' failures. """

    def __init__(
            self,
            msg: str,
            *,
            outcomes: dict[ids.HandlerId, execution.Outcome],
    ) -> None:
        super().__init__(msg)
        self.outcomes = outcomes


async def authenticator(
        *,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        indices: ephemera.Indices,
        vault: credentials.Vault,
        memo: ephemera.AnyMemo,
) -> NoReturn:
    """ Keep the credentials forever up to date. """
    counter: int = 0 if vault.is_empty() else 1
    while True:
        await authenticate(
            registry=registry,
            settings=settings,
            indices=indices,
            vault=vault,
            memo=memo,
            _activity_title="Re-authentication" if counter else "Initial authentication",
        )
        counter += 1


async def authenticate(
        *,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        indices: ephemera.Indices,
        vault: credentials.Vault,
        memo: ephemera.AnyMemo,
        _activity_title: str = "Authentication",
) -> None:
    """ Retrieve the credentials once, successfully or not, and exit. """
    # We do not need the locks protection here. There is only one activity for vault population.
    # Even with 2+ activities, if the vault is empty, all consumers will be blocked by waiting.
    # The API clients wake up only on the final population with the internal lock protection.

    # Sleep most of the time waiting for a signal to re-auth.
    await vault.wait_for_emptiness()

    # Log initial and re-authentications differently, for readability.
    logger.info(f"{_activity_title} has been initiated.")

    activity_results = await run_activity(
        lifecycle=lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        activity=causes.Activity.AUTHENTICATION,
        indices=indices,
        memo=memo,
    )

    if activity_results:
        logger.info(f"{_activity_title} has finished.")
    else:
        logger.warning(f"{_activity_title} has failed: "
                       f"no credentials were retrieved from the login handlers.")

    # Feed the credentials into the vault, and unfreeze the re-authenticating clients.
    await vault.populate({str(handler_id): info for handler_id, info in activity_results.items()})


async def run_activity(
        *,
        lifecycle: execution.LifeCycleFn,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        activity: causes.Activity,
        indices: ephemera.Indices,
        memo: ephemera.AnyMemo,
) -> dict[ids.HandlerId, execution.Result]:
    logger = logging.getLogger(f'kopf.activities.{activity.value}')

    # For the activity handlers, we have neither bodies, nor patches, just the state.
    cause = causes.ActivityCause(
        logger=logger,
        activity=activity,
        settings=settings,
        indices=indices,
        memo=memo,
    )
    handlers = registry._activities.get_handlers(activity=activity)
    state = progression.State.from_scratch().with_handlers(handlers)
    outcomes: dict[ids.HandlerId, execution.Outcome] = {}
    while not state.done:
        current_outcomes = await execution.execute_handlers_once(
            lifecycle=lifecycle,
            settings=settings,
            handlers=handlers,
            cause=cause,
            state=state,
        )
        outcomes |= current_outcomes
        state = state.with_outcomes(current_outcomes)
        await aiotime.sleep(state.delay)

    # Activities assume that all handlers must eventually succeed.
    # We raise from the 1st exception only: just to have something real in the tracebacks.
    # For multiple handlers' errors, the logs should be investigated instead.
    exceptions = [outcome.exception
                  for outcome in outcomes.values()
                  if outcome.exception is not None]
    if exceptions:
        raise ActivityError("One or more handlers failed.", outcomes=outcomes) from exceptions[0]

    # If nothing has failed, we return identifiable results. The outcomes/states are internal.
    # The order of results is not guaranteed (the handlers can succeed on one of the retries).
    results = {handler_id: outcome.result
               for handler_id, outcome in outcomes.items()
               if outcome.result is not None}
    return results



import asyncio

import pytest

import kopf
from kopf._core.reactor.running import spawn_tasks


@pytest.fixture(autouse=True)
def _safe_tasks(registry, settings):
    """Block startup to keep guarded tasks idle; disable SIGKILL scheduling."""
    settings.process.ultimate_exiting_timeout = None

    @kopf.on.startup(registry=registry)
    async def block_startup(**_):
        await asyncio.sleep(10)


async def _cancel_all(tasks):
    for task in tasks:
        task.cancel()
    await asyncio.gather(*tasks, return_exceptions=True)


# --- Argument validation ---


async def test_namespace_and_namespaces_cannot_coexist(registry, settings):
    with pytest.raises(TypeError, match="Either namespaces= or namespace= can be passed"):
        await spawn_tasks(
            registry=registry,
            settings=settings,
            namespaces=['ns1'],
            namespace='ns2',
        )


async def test_clusterwide_and_namespaces_cannot_coexist(registry, settings):
    with pytest.raises(TypeError, match="cluster-wide or namespaced, not both"):
        await spawn_tasks(
            registry=registry,
            settings=settings,
            clusterwide=True,
            namespaces=['ns1'],
        )


async def test_deprecated_namespace_kwarg(registry, settings):
    with pytest.warns(DeprecationWarning, match="namespace= is deprecated"):
        tasks = await spawn_tasks(
            registry=registry,
            settings=settings,
            namespace='ns1',
        )
    await _cancel_all(tasks)


async def test_missing_scope_defaults_to_clusterwide(registry, settings):
    with pytest.warns(FutureWarning, match="Absence of either namespaces or cluster-wide"):
        tasks = await spawn_tasks(
            registry=registry,
            settings=settings,
        )
    assert settings.peering.clusterwide is True
    await _cancel_all(tasks)


# --- Settings mapping ---


async def test_clusterwide_mapped_to_settings(registry, settings):
    assert settings.peering.clusterwide is False
    tasks = await spawn_tasks(
        registry=registry,
        settings=settings,
        clusterwide=True,
    )
    assert settings.peering.clusterwide is True
    await _cancel_all(tasks)


async def test_peering_name_mapped_to_settings(registry, settings):
    assert settings.peering.mandatory is False
    tasks = await spawn_tasks(
        registry=registry,
        settings=settings,
        clusterwide=True,
        peering_name='my-peering',
    )
    assert settings.peering.mandatory is True
    assert settings.peering.name == 'my-peering'
    await _cancel_all(tasks)


async def test_standalone_mapped_to_settings(registry, settings):
    assert settings.peering.standalone is False
    tasks = await spawn_tasks(
        registry=registry,
        settings=settings,
        clusterwide=True,
        standalone=True,
    )
    assert settings.peering.standalone is True
    await _cancel_all(tasks)


async def test_priority_mapped_to_settings(registry, settings):
    assert settings.peering.priority == 0
    tasks = await spawn_tasks(
        registry=registry,
        settings=settings,
        clusterwide=True,
        priority=100,
    )
    assert settings.peering.priority == 100
    await _cancel_all(tasks)


# --- Task composition ---


async def test_always_present_tasks(registry, settings):
    tasks = await spawn_tasks(
        registry=registry,
        settings=settings,
        clusterwide=True,
    )
    task_names = {t.get_name() for t in tasks}

    assert task_names == {
        "stop-flag checker",
        "ultimate termination",
        "startup/cleanup activities",
        "daemon killer",
        "poster of events",
        "admission insights chain",
        "admission validating configuration manager",
        "admission mutating configuration manager",
        "admission webhook server",
        "resource observer",
        "namespace observer",
        "multidimensional multitasker",
    }
    assert "health reporter" not in task_names

    await _cancel_all(tasks)


async def test_liveness_endpoint_adds_health_reporter(registry, settings):
    tasks = await spawn_tasks(
        registry=registry,
        settings=settings,
        clusterwide=True,
        liveness_endpoint="/health",
    )
    task_names = {t.get_name() for t in tasks}
    assert "health reporter" in task_names
    await _cancel_all(tasks)


# The cancellation of "core" tasks is tested elsewhere. We just test that it is passed, not lost.
async def test_core_tasks_passed_to_activities(registry, settings, mocker):
    mock = mocker.patch('kopf._core.reactor.running.startup_cleanup_activities')
    tasks = await spawn_tasks(
        registry=registry,
        settings=settings,
        clusterwide=True,
    )
    core_tasks = mock.call_args.kwargs['core_tasks']
    core_task_names = {t.get_name() for t in core_tasks}
    assert core_task_names == {"credentials retriever"}
    await _cancel_all(tasks)
    await _cancel_all(core_tasks)  # because not awaited and not stopped in the mock


async def test_command_replaces_orchestrator(registry, settings):
    async def my_command():
        await asyncio.Event().wait()

    tasks = await spawn_tasks(
        registry=registry,
        settings=settings,
        clusterwide=True,
        _command=my_command(),
    )
    task_names = {t.get_name() for t in tasks}
    assert "the command" in task_names
    assert "multidimensional multitasker" not in task_names
    await _cancel_all(tasks)



"""
All the functions to calculate the diffs of the dicts.
"""
import collections.abc
import enum
from collections.abc import Iterable, Iterator, Sequence
from typing import Any, NamedTuple, overload

from kopf._cogs.structs import dicts


class DiffScope(enum.Flag):
    """
    Scope limitation for the diffs' fields to be noticed or ignored.

    In the full-scoped diff (the default), both objects (diff source & target)
    are treated equally important, and the diff is calculated from the left
    to the right one for all fields.

    In the left-scoped diff, only the left object (diff source) is considered
    important, and only the differences for the fields found in the left object
    (source) are checked. Extra fields in the right object (target) are ignored.

    In the right-scoped diff, only the fields in the right object (diff target)
    are scanned. Extra fields in the left object (diff source) are ignored.
    """
    RIGHT = enum.auto()
    LEFT = enum.auto()
    FULL = LEFT | RIGHT


class DiffOperation(str, enum.Enum):
    ADD = 'add'
    CHANGE = 'change'
    REMOVE = 'remove'

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return repr(self.value)


class DiffItem(NamedTuple):
    operation: DiffOperation
    field: dicts.FieldPath
    old: Any
    new: Any

    def __repr__(self) -> str:
        return repr(tuple(self))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, collections.abc.Sequence):
            return tuple(self) == tuple(other)
        else:
            return NotImplemented

    def __ne__(self, other: object) -> bool:
        if isinstance(other, collections.abc.Sequence):
            return tuple(self) != tuple(other)
        else:
            return NotImplemented

    @property
    def op(self) -> DiffOperation:
        return self.operation


class Diff(Sequence[DiffItem]):

    def __init__(self, __items: Iterable[DiffItem]):
        super().__init__()
        self._items = tuple(DiffItem(*item) for item in __items)

    def __hash__(self) -> int:
        # Hashes mark diffs as immutable to be usable as dataclasses' defaults in Python 3.11.
        return hash(self._items)

    def __repr__(self) -> str:
        return repr(self._items)

    def __len__(self) -> int:
        return len(self._items)

    def __iter__(self) -> Iterator[DiffItem]:
        return iter(self._items)

    @overload
    def __getitem__(self, i: int) -> DiffItem:
        ...

    @overload
    def __getitem__(self, s: slice) -> Sequence[DiffItem]:
        ...

    def __getitem__(self, item: int | slice) -> DiffItem | Sequence[DiffItem]:
        return self._items[item]

    def __eq__(self, other: object) -> bool:
        if isinstance(other, collections.abc.Sequence):
            return tuple(self) == tuple(other)
        else:
            return NotImplemented

    def __ne__(self, other: object) -> bool:
        if isinstance(other, collections.abc.Sequence):
            return tuple(self) != tuple(other)
        else:
            return NotImplemented


def reduce_iter(
        d: Diff,
        path: dicts.FieldPath,
) -> Iterator[DiffItem]:
    for op, field, old, new in d:

        # As-is diff (i.e. a root field).
        if not path:
            yield DiffItem(op, tuple(field), old, new)

        # The diff-field is longer than the path: get "spec.struct" when "spec.struct.field" is set.
        # Retranslate the diff with the field prefix shrinked.
        elif tuple(field[:len(path)]) == tuple(path):
            yield DiffItem(op, tuple(field[len(path):]), old, new)

        # The diff-field is shorter than the path: get "spec.struct" when "spec={...}" is added.
        # Generate a new diff, with new ops, for the resolved sub-field.
        elif tuple(field) == tuple(path[:len(field)]):
            tail = path[len(field):]
            old_tail = dicts.resolve(old, tail, default=None)
            new_tail = dicts.resolve(new, tail, default=None)
            yield from diff_iter(old_tail, new_tail)


def reduce(
        d: Diff,
        path: dicts.FieldPath,
) -> Diff:
    return Diff(reduce_iter(d, path))


def diff_iter(
        a: Any,
        b: Any,
        path: dicts.FieldPath = (),
        *,
        scope: DiffScope = DiffScope.FULL,
) -> Iterator[DiffItem]:
    """
    Calculate the diff between two dicts.

    Yields the tuple of form ``(op, path, old, new)``,
    where ``op`` is either ``"add"``/``"change"``/``"remove"``,
    ``path`` is a tuple with the field names (empty tuple means root),
    and the ``old`` & ``new`` values (`None` for addition/removal).

    List values are treated as a whole, and not recursed into.
    Therefore, an addition/removal of a list item is considered
    as a change of the whole value.

    If the deep diff for lists/sets is needed, see the libraries:

    * https://dictdiffer.readthedocs.io/en/latest/
    * https://github.com/seperman/deepdiff
    * https://python-json-patch.readthedocs.io/en/latest/tutorial.html
    """
    match a, b:
        case a, b if a == b:  # incl. cases when both are None
            pass
        case None, _:
            yield DiffItem(DiffOperation.ADD, path, a, b)
        case _, None:
            yield DiffItem(DiffOperation.REMOVE, path, a, b)
        case collections.abc.Mapping(), collections.abc.Mapping():
            a_keys = frozenset(a.keys())
            b_keys = frozenset(b.keys())
            for key in (b_keys - a_keys if DiffScope.RIGHT in scope else ()):
                yield from diff_iter(None, b[key], path=path+(key,), scope=scope)
            for key in (a_keys - b_keys if DiffScope.LEFT in scope else ()):
                yield from diff_iter(a[key], None, path=path+(key,), scope=scope)
            for key in (a_keys & b_keys):
                yield from diff_iter(a[key], b[key], path=path+(key,), scope=scope)
        case _:
            yield DiffItem(DiffOperation.CHANGE, path, a, b)


def diff(
        a: Any,
        b: Any,
        path: dicts.FieldPath = (),
        *,
        scope: DiffScope = DiffScope.FULL,
) -> Diff:
    """
    Same as ``diff_iter``, but returns the collection instead of an iterator.
    """
    return Diff(diff_iter(a, b, path=path, scope=scope))


EMPTY = diff(None, None)



"""
Engines are things that run around the reactor (see :mod:`kopf._core.reactor`)
to help it function at full strength, but are not part of it.
For example, all ever-running side tasks for peering and k8s-event-posting.

The reactor and engines exchange state with each other
via the provided synchronization objects, usually asyncio events & queues.
"""
